from pathlib import Path
from types import MappingProxyType
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    Iterable,
    Optional,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from pydantic import BaseModel, ConfigDict

if TYPE_CHECKING:
    from typing_extensions import Self

    ReprArgs = Iterable[tuple[str | None, Any]]

__all__ = ["UseqModel", "FrozenModel"]

_T = TypeVar("_T", bound="FrozenModel")
_Y = TypeVar("_Y", bound="UseqModel")


def _non_default_repr_args(obj: BaseModel, fields: "ReprArgs") -> "ReprArgs":
    """Set fields on a model instance."""
    return [
        (k, val)
        for k, val in fields
        if k in obj.model_fields
        and val
        != (
            factory()
            if (factory := obj.model_fields[k].default_factory) is not None
            else obj.model_fields[k].default
        )
    ]


# TODO: consider removing this and using model_copy directly
class _ReplaceableModel(BaseModel):
    def replace(self, **kwargs: Any) -> "Self":
        """Return a new instance replacing specified kwargs with new values.

        This model is immutable, so this method is useful for creating a new
        sequence with only a few fields changed.  The uid of the new sequence will
        be different from the original.

        The difference between this and `self.model_copy(update={...})` is that this
        method will perform validation and casting on the new values, whereas `copy`
        assumes that all objects are valid and will not perform any validation or
        casting.
        """
        # only get values for top level fields
        d = {k: getattr(self, k) for k in self.model_fields if k != "uid"}
        return type(self).model_validate({**d, **kwargs})

    def __repr_args__(self) -> "ReprArgs":
        """Only show fields that are not None or equal to their default value."""
        return _non_default_repr_args(self, super().__repr_args__())


class FrozenModel(_ReplaceableModel):
    model_config: ClassVar["ConfigDict"] = ConfigDict(
        populate_by_name=True,
        extra="ignore",
        frozen=True,
        json_encoders={MappingProxyType: dict},
    )


class MutableModel(_ReplaceableModel):
    model_config: ClassVar["ConfigDict"] = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        validate_default=True,
        extra="ignore",
    )


class UseqModel(FrozenModel):
    @classmethod
    def from_file(cls: Type[_Y], path: Union[str, Path]) -> _Y:
        """Return an instance of this class from a file.  Supports JSON and YAML."""
        path = Path(path)
        if path.suffix in {".yaml", ".yml"}:
            import yaml

            obj = yaml.safe_load(path.read_bytes())
        elif path.suffix == ".json":
            return cls.model_validate_json(path.read_bytes())
        else:  # pragma: no cover
            raise ValueError(f"Unknown file type: {path.suffix}")

        return cls.model_validate(obj)

    def yaml(
        self,
        *,
        include: Optional[Union[set, dict]] = None,
        exclude: Optional[Union[set, dict]] = None,
        by_alias: bool = False,
        exclude_unset: bool = True,  # pydantic has False by default
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        stream: Optional[IO[str]] = None,
    ) -> Optional[str]:
        """Generate a YAML representation of the model.

        Returns
        -------
        yaml : str or None
            YAML output ... If `stream` is provided, returns `None`.
        """
        from datetime import timedelta
        from enum import Enum

        import yaml

        yaml.SafeDumper.add_multi_representer(
            timedelta, lambda dumper, data: dumper.represent_str(str(data))
        )
        yaml.SafeDumper.add_multi_representer(
            Enum, lambda dumper, data: dumper.represent_str(str(data.value))
        )
        yaml.SafeDumper.add_multi_representer(
            MappingProxyType, lambda dumper, data: dumper.represent_dict(data)
        )
        yaml.SafeDumper.add_multi_representer(
            np.floating, lambda dumper, d: dumper.represent_float(float(d))
        )

        data = self.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return yaml.safe_dump(data, stream=stream)



from __future__ import annotations

from typing import Iterable

import pytest

import useq
from useq import AxesBasedAF, HardwareAutofocus, MDASequence

ZRANGE2 = useq.ZRangeAround(range=2, step=1)
ZPOS_30 = useq.Position(z=30.0)
ZPOS_200 = useq.Position(z=200.0)
GRID_PLAN = useq.GridRelative(rows=2, columns=1)
TWO_CH = MDASequence(stage_positions=[ZPOS_30], channels=["DAPI", "FITC"])
TWO_CH_TWO_P = TWO_CH.replace(stage_positions=[ZPOS_30, ZPOS_200])
TWO_CH_TWO_P_GRID = TWO_CH_TWO_P.replace(grid_plan=GRID_PLAN)
TWO_CH_TWO_P_T = TWO_CH_TWO_P.replace(time_plan={"interval": 1, "loops": 2})
NO_CH_ZSTACK = MDASequence(stage_positions=[ZPOS_30], z_plan=ZRANGE2)
TWO_CH_ZSTACK = NO_CH_ZSTACK.replace(channels=["DAPI", "FITC"])
AF = AxesBasedAF(autofocus_device_name="Z", autofocus_motor_offset=40, axes=())
AF_C = AF.model_copy(update={"axes": ("c",)})
AF_G = AF.model_copy(update={"axes": ("g",)})
AF_P = AF.model_copy(update={"axes": ("p",)})
AF_Z = AF.model_copy(update={"axes": ("z",)})
AF_SEQ_C = MDASequence(autofocus_plan=AF_C)
SUB_P_AF_C = useq.Position(z=10, sequence=AF_SEQ_C)
SUB_P_AF_G = useq.Position(z=10, sequence=MDASequence(autofocus_plan=AF_G))
SUB_P_AF_P = useq.Position(z=10, sequence=MDASequence(autofocus_plan=AF_P))
SUB_P_AF_Z = useq.Position(z=10, sequence=MDASequence(autofocus_plan=AF_Z))
TWO_CH_SUBPAF_C = TWO_CH.replace(stage_positions=[ZPOS_30, SUB_P_AF_C])
TWO_CH_SUBPAF_Z = TWO_CH.replace(stage_positions=[ZPOS_30, SUB_P_AF_Z])

# fmt: off
AF_TESTS: list[tuple[MDASequence, tuple[str, ...], Iterable[int]]] = [
    (NO_CH_ZSTACK, ("c",), ()),
    (TWO_CH_ZSTACK, ("c",), (0, 4)),
    (TWO_CH_ZSTACK, ("z",), range(0, 11, 2)),
    (TWO_CH_ZSTACK.replace(axis_order="tpgzc"), ("c",), range(0, 11, 2)),
    (TWO_CH_ZSTACK.replace(axis_order="tpgzc"), ("z",), (0, 3, 6)),
    (TWO_CH_ZSTACK, ("z",), range(0, 11, 2)),
    (TWO_CH.replace(grid_plan=GRID_PLAN), ("g",), (0, 3)),
    (TWO_CH_TWO_P, ("g",), ()),
    (TWO_CH_TWO_P_GRID, ("g",), range(0, 10, 3)),
    (TWO_CH_TWO_P_GRID, ("p",), (0, 5)),
    (TWO_CH_TWO_P_T, ("t",), (0, 5)),
    (TWO_CH_TWO_P_T, ("t", "p"), range(0, 10, 3)),
    (TWO_CH_SUBPAF_C, (), (2, 4)),
    (TWO_CH.replace(stage_positions=[ZPOS_30, SUB_P_AF_G]), (), ()),
    (TWO_CH.replace(stage_positions=[ZPOS_30, SUB_P_AF_G], grid_plan=GRID_PLAN), (), (4, 7)),
    (TWO_CH.replace(stage_positions=[ZPOS_30, SUB_P_AF_P], grid_plan=GRID_PLAN), (), (4,)),
    (TWO_CH_SUBPAF_C.replace(grid_plan=GRID_PLAN), (), range(4, 11, 2)),
    (TWO_CH_SUBPAF_C.replace(z_plan=ZRANGE2), (), (6, 10)),
    (TWO_CH_SUBPAF_Z.replace(z_plan=ZRANGE2), (), range(6, 17, 2)),
    (TWO_CH_SUBPAF_Z.replace(z_plan=ZRANGE2), ("p",), (0, *tuple(range(7, 18, 2)))),
    (TWO_CH_SUBPAF_C.replace(z_plan=ZRANGE2, grid_plan=GRID_PLAN), ("c",), range(0, 29, 4)),
    (TWO_CH.replace(stage_positions=[ZPOS_30, useq.Position(z=10, sequence=MDASequence(autofocus_plan=AF_C, grid_plan=GRID_PLAN))]), (), (2, 5)),
    (TWO_CH.replace(stage_positions=[SUB_P_AF_C, useq.Position(z=10, sequence=MDASequence(autofocus_plan=AF_G, grid_plan=GRID_PLAN))]), (), range(0, 11, 2)),
    (TWO_CH.replace(stage_positions=[ZPOS_200, useq.Position(z=10, sequence=MDASequence(z_plan=ZRANGE2))]), ("z",), range(2, 13, 2)),
    (TWO_CH.replace(stage_positions=[ZPOS_200, useq.Position(z=10, sequence=MDASequence(z_plan=ZRANGE2))]), ("c",), (0, 2, 4, 8)),
    (TWO_CH.replace(stage_positions=[ZPOS_200, useq.Position(z=10, sequence=MDASequence(z_plan=ZRANGE2))]), (), ()),
]
# fmt: on


@pytest.mark.parametrize("mda, af_axes, expected_af_indices", AF_TESTS)
def test_autofocus(
    mda: MDASequence,
    af_axes: tuple[str, ...],
    expected_af_indices: Iterable[int],
) -> None:
    if af_axes:
        mda = mda.replace(autofocus_plan=AF.model_copy(update={"axes": af_axes}))
        assert isinstance(mda.autofocus_plan, AxesBasedAF)
        assert mda.autofocus_plan.axes == af_axes

    actual = [i for i, e in enumerate(mda) if isinstance(e.action, HardwareAutofocus)]
    expected = list(expected_af_indices)
    assert expected == actual, "Unexpected AF indices"


def test_autofocus_z_pos() -> None:
    mda = TWO_CH.replace(
        stage_positions=[ZPOS_200], z_plan=ZRANGE2, autofocus_plan=AF_C
    )
    assert all(e.z_pos == 200 for e in mda if isinstance(e.action, HardwareAutofocus))


def test_autofocus_z_pos_abovebelow() -> None:
    mda = TWO_CH.replace(
        stage_positions=[ZPOS_200],
        z_plan=useq.ZAboveBelow(above=2, below=2, step=1),
        autofocus_plan=AF_C,
    )
    assert all(e.z_pos == 200 for e in mda if isinstance(e.action, HardwareAutofocus))


def test_autofocus_z_pos_af_sub_sequence() -> None:
    mda = TWO_CH.replace(stage_positions=[SUB_P_AF_C], z_plan=ZRANGE2)
    assert all(e.z_pos == 10 for e in mda if isinstance(e.action, HardwareAutofocus))


def test_autofocus_z_pos_af_sub_sequence_zplan() -> None:
    mda = TWO_CH_SUBPAF_C
    assert all(e.z_pos == 10 for e in mda if isinstance(e.action, HardwareAutofocus))


def test_autofocus_z_pos_multi_plans() -> None:
    mda = TWO_CH.replace(
        stage_positions=[ZPOS_200],
        grid_plan=GRID_PLAN,
        z_plan=ZRANGE2,
        autofocus_plan=AF_C,
    )

    assert all(e.z_pos == 200 for e in mda if isinstance(e.action, HardwareAutofocus))


def test_af_no_name() -> None:
    list(
        MDASequence(
            time_plan={"interval": 1, "loops": 2},
            autofocus_plan=AxesBasedAF(axes=("t", "c")),
        )
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable

try:
    import matplotlib.pyplot as plt
    from matplotlib import patches
except ImportError as e:
    raise ImportError(
        "Matplotlib is required for plotting functions.  Please install matplotlib."
    ) from e

if TYPE_CHECKING:
    from matplotlib.axes import Axes

    from useq._plate import WellPlatePlan
    from useq._position import PositionBase


def plot_points(points: Iterable[PositionBase], ax: Axes | None = None) -> None:
    """Plot a list of positions.

    Can be used with any iterable of PositionBase objects.
    """
    if ax is None:
        _, ax = plt.subplots()

    x, y = zip(*[(point.x, point.y) for point in points])
    ax.scatter(x, y)
    ax.scatter(x[0], y[0], color="red")  # mark the first point
    ax.plot(x, y, alpha=0.5, color="gray")  # connect the points
    ax.axis("equal")
    plt.show()


def plot_plate(
    plate_plan: WellPlatePlan, show_axis: bool = True, ax: Axes | None = None
) -> None:
    if ax is None:
        _, ax = plt.subplots()

    # hide axes
    if not show_axis:
        ax.axis("off")

    # ################ draw outline of all wells ################
    height, width = plate_plan.plate.well_size  # mm
    height, width = height * 1000, width * 1000  # µm

    kwargs = {}
    offset_x, offset_y = 0.0, 0.0
    if plate_plan.plate.circular_wells:
        patch_type: Callable = patches.Ellipse
    else:
        patch_type = patches.Rectangle
        offset_x, offset_y = -width / 2, -height / 2
        kwargs["rotation_point"] = "center"

    for well in plate_plan.all_well_positions:
        sh = patch_type(
            (well.x + offset_x, well.y + offset_y),  # type: ignore[operator]
            width=width,
            height=height,
            angle=plate_plan.rotation or 0,
            facecolor="none",
            edgecolor="gray",
            linewidth=0.5,
            linestyle="--",
            **kwargs,
        )
        ax.add_patch(sh)

    ################ plot image positions ################
    w, h = plate_plan.well_points_plan.fov_width, plate_plan.well_points_plan.fov_height

    for img_point in plate_plan.image_positions:
        x, y = float(img_point.x), float(img_point.y)  # type: ignore[arg-type] # µm
        if w and h:
            ax.add_patch(
                patches.Rectangle(
                    (x - w / 2, y - h / 2),
                    width=w,
                    height=h,
                    facecolor="magenta",
                    edgecolor="gray",
                    linewidth=0.5,
                    alpha=0.5,
                )
            )
        else:
            plt.plot(x, y, "mo", markersize=3, alpha=0.5)

    # ################ draw names on used wells ################
    offset_x, offset_y = -width / 2, -height / 2
    for well in plate_plan.selected_well_positions:
        x, y = float(well.x), float(well.y)  # type: ignore[arg-type]
        # draw name next to spot
        ax.text(x + offset_x, y - offset_y, well.name, fontsize=7)

    ax.axis("equal")
    plt.show()



from __future__ import annotations

from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    List,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

import numpy as np
from annotated_types import Gt  # noqa: TCH002
from pydantic import (
    Field,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated

from useq._base_model import FrozenModel, UseqModel
from useq._grid import RandomPoints, RelativeMultiPointPlan, Shape
from useq._plate_registry import _PLATE_REGISTRY
from useq._position import Position, PositionBase, RelativePosition

if TYPE_CHECKING:
    from pydantic_core import core_schema

    Index = Union[int, List[int], slice]
    IndexExpression = Union[Tuple[Index, ...], Index]


class WellPlate(FrozenModel):
    """A multi-well plate definition.

    Parameters
    ----------
    rows : int
        The number of rows in the plate. Must be > 0.
    columns : int
        The number of columns in the plate. Must be > 0.
    well_spacing : tuple[float, float] | float
        The center-to-center distance in mm (pitch) between wells in the x and y
        directions. If a single value is provided, it is used for both x and y.
    well_size : tuple[float, float] | float
        The size in mm of each well in the x and y directions. If the well is squared or
        rectangular, this is the width and height of the well. If the well is circular,
        this is the diameter. If a single value is provided, it is used for both x and
        y.
    circular_wells : bool
        Whether wells are circular (True) or squared/rectangular (False).
    name : str
        A name for the plate.
    """

    rows: Annotated[int, Gt(0)]
    columns: Annotated[int, Gt(0)]
    well_spacing: Tuple[float, float]  # (x, y)
    well_size: Tuple[float, float]  # (width, height)
    circular_wells: bool = True
    name: str = ""

    @property
    def size(self) -> int:
        """Return the total number of wells."""
        return self.rows * self.columns

    @property
    def shape(self) -> tuple[int, int]:
        """Return the shape of the plate."""
        return self.rows, self.columns

    @property
    def all_well_indices(self) -> np.ndarray:
        """Return the indices of all wells as array with shape (Rows, Cols, 2)."""
        Y, X = np.meshgrid(np.arange(self.rows), np.arange(self.columns), indexing="ij")
        return np.stack([Y, X], axis=-1)

    def indices(self, expr: IndexExpression) -> np.ndarray:
        """Return the indices for any index expression as array with shape (N, 2)."""
        return self.all_well_indices[expr].reshape(-1, 2).T

    @property
    def all_well_names(self) -> np.ndarray:
        """Return the names of all wells as array of strings with shape (Rows, Cols)."""
        return np.array(
            [
                [f"{_index_to_row_name(r)}{c+1}" for c in range(self.columns)]
                for r in range(self.rows)
            ]
        )

    @field_validator("well_spacing", "well_size", mode="before")
    def _validate_well_spacing_and_size(cls, value: Any) -> Any:
        return (value, value) if isinstance(value, (int, float)) else value

    @model_validator(mode="before")
    @classmethod
    def validate_plate(cls, value: Any) -> Any:
        if isinstance(value, (int, float)):
            value = f"{int(value)}-well"
        return cls.from_str(value) if isinstance(value, str) else value

    @classmethod
    def from_str(cls, name: str) -> WellPlate:
        """Lookup a plate by registered name.

        Use `useq.register_well_plates` to add new plates to the registry.
        """
        try:
            obj = _PLATE_REGISTRY[name]
        except KeyError as e:
            raise ValueError(
                f"Unknown plate name {name!r}. "
                "Use `useq.register_well_plates` to add new plate definitions"
            ) from e
        if isinstance(obj, dict) and "name" not in obj:
            obj = {**obj, "name": name}
        return WellPlate.model_validate(obj)


class WellPlatePlan(UseqModel, Sequence[Position]):
    """A plan for acquiring images from a multi-well plate.

    Parameters
    ----------
    plate : WellPlate | str | int
        The well-plate definition. Minimally including rows, columns, and well spacing.
        If expressed as a string, it is assumed to be a key in
        `useq.registered_well_plate_keys`.
    a1_center_xy : tuple[float, float]
        The stage coordinates in µm of the center of well A1 (top-left corner).
    rotation : float | None
        The rotation angle in degrees (anti-clockwise) of the plate.
        If None, no rotation is applied.
        If expressed as a string, it is assumed to be an angle with units (e.g., "5°",
        "4 rad", "4.5deg").
        If expressed as an arraylike, it is assumed to be a 2x2 rotation matrix
        `[[cos, -sin], [sin, cos]]`, or a 4-tuple `(cos, -sin, sin, cos)`.
    selected_wells : IndexExpression | None
        Any <=2-dimensional index expression for selecting wells.
        for example:
        -   None -> No wells are selected.
        -   slice(0) -> (also) select no wells.
        -   slice(None) -> Selects all wells.
        -   0 -> Selects the first row.
        -   [0, 1, 2] -> Selects the first three rows.
        -   slice(1, 5) -> selects wells from row 1 to row 4.
        -   (2, slice(1, 4)) -> select wells in the second row and only columns 1 to 3.
        -   ([1, 2], [3, 4]) -> select wells in (row, column): (1, 3) and (2, 4)
    well_points_plan : GridRowsColumns | RandomPoints | Position
        A plan for acquiring images within each well. This can be a single position
        (for a single image per well), a GridRowsColumns (for a grid of images),
        or RandomPoints (for random points within each well).
    """

    plate: WellPlate
    a1_center_xy: Tuple[float, float]
    rotation: Union[float, None] = None
    selected_wells: Union[Tuple[Tuple[int, ...], Tuple[int, ...]], None] = None
    well_points_plan: RelativeMultiPointPlan = Field(
        default_factory=RelativePosition, union_mode="left_to_right"
    )

    def __repr_args__(self) -> Iterable[Tuple[str | None, Any]]:
        for item in super().__repr_args__():
            if item[0] == "selected_wells":
                # improve repr for selected_wells
                yield "selected_wells", _expression_repr(item[1])
            else:
                yield item

    @field_validator("plate", mode="before")
    @classmethod
    def _validate_plate(cls, value: Any) -> Any:
        return WellPlate.validate_plate(value)  # type: ignore [operator]

    @field_validator("well_points_plan", mode="wrap")
    @classmethod
    def _validate_well_points_plan(
        cls,
        value: Any,
        handler: core_schema.ValidatorFunctionWrapHandler,
        info: core_schema.ValidationInfo,
    ) -> Any:
        value = handler(value)
        if plate := info.data.get("plate"):
            if isinstance(value, RandomPoints):
                plate = cast(WellPlate, plate)
                kwargs = value.model_dump(mode="python")
                if value.max_width == np.inf:
                    well_size_x = plate.well_size[0] * 1000  # convert to µm
                    kwargs["max_width"] = well_size_x - (value.fov_width or 0.1)
                if value.max_height == np.inf:
                    well_size_y = plate.well_size[1] * 1000  # convert to µm
                    kwargs["max_height"] = well_size_y - (value.fov_height or 0.1)
                if "shape" not in value.__pydantic_fields_set__:
                    kwargs["shape"] = (
                        Shape.ELLIPSE if plate.circular_wells else Shape.RECTANGLE
                    )
                value = RandomPoints(**kwargs)
        return value

    @field_validator("rotation", mode="before")
    @classmethod
    def _validate_rotation(cls, value: Any) -> Any:
        if isinstance(value, str):
            # assume string representation of an angle
            # infer deg or radians from the string
            if "rad" in value:
                value = value.replace("rad", "").strip()
                # convert to degrees
                return np.degrees(float(value))
            if "°" in value or "˚" in value or "deg" in value:
                value = value.replace("°", "").replace("˚", "").replace("deg", "")
                return float(value.strip())
        if isinstance(value, (tuple, list)):
            ary = np.array(value).flatten()
            if len(ary) != 4:  # pragma: no cover
                raise ValueError("Rotation matrix must have 4 elements")
            # convert (cos, -sin, sin, cos) to angle in degrees, anti-clockwise
            return np.degrees(np.arctan2(ary[2], ary[0]))
        return value

    @field_validator("selected_wells", mode="wrap")
    @classmethod
    def _validate_selected_wells(
        cls, value: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
    ) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
        plate = info.data.get("plate")
        if not isinstance(plate, WellPlate):
            raise ValueError("Plate must be defined before selecting wells")

        if isinstance(value, list):
            value = tuple(value)
        # make falsey values select no wells (rather than all wells)
        if not value:
            value = slice(0)

        try:
            selected = plate.indices(value)
        except (TypeError, IndexError) as e:
            raise ValueError(
                f"Invalid well selection {value!r} for plate of "
                f"shape {plate.shape}: {e}"
            ) from e

        return handler(selected)  # type: ignore [no-any-return]

    @property
    def rotation_matrix(self) -> np.ndarray:
        """Convert self.rotation (which is in degrees) to a rotation matrix."""
        if self.rotation is None:
            return np.eye(2)
        rads = np.radians(self.rotation)
        return np.array([[np.cos(rads), np.sin(rads)], [-np.sin(rads), np.cos(rads)]])

    def __iter__(self) -> Iterator[Position]:  # type: ignore
        """Iterate over the selected positions."""
        yield from self.image_positions

    def __len__(self) -> int:
        """Return the total number of points (stage positions) to be acquired."""
        if self.selected_wells is None:
            n_wells = 0
        else:
            n_wells = len(self.selected_wells[0])
        return n_wells * self.num_points_per_well

    def __bool__(self) -> bool:
        """bool(WellPlatePlan) == True."""
        return True

    @overload
    def __getitem__(self, index: int) -> Position: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Position]: ...

    def __getitem__(self, index: int | slice) -> Position | Sequence[Position]:
        """Return the selected position(s) at the given index."""
        return self.image_positions[index]

    @property
    def num_points_per_well(self) -> int:
        """Return the number of points per well."""
        if isinstance(self.well_points_plan, PositionBase):
            return 1
        else:
            return self.well_points_plan.num_positions()

    @property
    def all_well_indices(self) -> np.ndarray:
        """Return the indices of all wells as array with shape (Rows, Cols, 2)."""
        return self.plate.all_well_indices

    @property
    def selected_well_indices(self) -> np.ndarray:
        """Return the indices of selected wells as array with shape (N, 2)."""
        return self.plate.all_well_indices[self.selected_wells].reshape(-1, 2)

    @cached_property
    def all_well_coordinates(self) -> np.ndarray:
        """Return the stage coordinates of all wells as array with shape (N, 2)."""
        return self._transorm_coords(self.plate.all_well_indices.reshape(-1, 2))

    @cached_property
    def selected_well_coordinates(self) -> np.ndarray:
        """Return the stage coordinates of selected wells as array with shape (N, 2)."""
        return self._transorm_coords(self.selected_well_indices)

    @property
    def all_well_names(self) -> np.ndarray:
        """Return the names of all wells as array of strings with shape (Rows, Cols)."""
        return self.plate.all_well_names

    @property
    def selected_well_names(self) -> list[str]:
        """Return the names of selected wells."""
        return list(self.all_well_names[self.selected_wells].reshape(-1))

    def _transorm_coords(self, coords: np.ndarray) -> np.ndarray:
        """Transform coordinates to the plate coordinate system."""
        # create homogenous coordinates
        h_coords = np.column_stack((coords, np.ones(coords.shape[0])))
        # transform
        transformed = self.affine_transform @ h_coords.T
        # strip homogenous coordinate
        return (transformed[:2].T).reshape(coords.shape)  # type: ignore[no-any-return]

    @property
    def all_well_positions(self) -> Sequence[Position]:
        """Return all wells (centers) as Position objects."""
        return [
            Position(x=x * 1000, y=y * 1000, name=name)  # convert to µm
            for (y, x), name in zip(
                self.all_well_coordinates, self.all_well_names.reshape(-1)
            )
        ]

    @cached_property
    def selected_well_positions(self) -> Sequence[Position]:
        """Return selected wells (centers) as Position objects."""
        return [
            Position(x=x * 1000, y=y * 1000, name=name)  # convert to µm
            for (y, x), name in zip(
                self.selected_well_coordinates, self.selected_well_names
            )
        ]

    @cached_property
    def image_positions(self) -> Sequence[Position]:
        """All image positions.

        This includes *both* selected wells and the image positions within each well
        based on the `well_points_plan`.  This is the primary property that gets used
        when iterating over the plan.
        """
        wpp = self.well_points_plan
        offsets = [wpp] if isinstance(wpp, RelativePosition) else wpp
        pos: List[Position] = []
        for well in self.selected_well_positions:
            pos.extend(well + offset for offset in offsets)
        return pos

    @property
    def affine_transform(self) -> np.ndarray:
        """Return transformation matrix that maps well indices to stage coordinates.

        This includes:
        1. scaling by plate.well_spacing
        2. rotation by rotation_matrix
        3. translation to a1_center_xy

        Note that the Y axis scale is inverted to go from linearly increasing index
        coordinates to cartesian "plate" coordinates (where y position decreases with
        increasing index.
        """
        translation = np.eye(3)
        a1_center_xy_mm = np.array(self.a1_center_xy) / 1000  # convert to mm
        translation[:2, 2] = a1_center_xy_mm[::-1]

        rotation = np.eye(3)
        rotation[:2, :2] = self.rotation_matrix

        scaling = np.eye(3)
        # invert the Y axis to convert "index" to "plate" coordinates.
        scale_x, scale_y = self.plate.well_spacing
        scaling[:2, :2] = np.diag([-scale_y, scale_x])

        return translation @ rotation @ scaling

    def plot(self, show_axis: bool = True) -> None:
        """Plot the selected positions on the plate."""
        from useq._plot import plot_plate

        plot_plate(self, show_axis=show_axis)


def _index_to_row_name(index: int) -> str:
    """Convert a zero-based column index to row name (A, B, ..., Z, AA, AB, ...)."""
    name = ""
    while index >= 0:
        name = chr(index % 26 + 65) + name
        index = index // 26 - 1
    return name


def _find_pattern(seq: Sequence[int]) -> tuple[list[int] | None, int | None]:
    n = len(seq)

    # Try different lengths of the potential repeating pattern
    for pattern_length in range(1, n // 2 + 1):
        pattern = list(seq[:pattern_length])
        repetitions = n // pattern_length

        # Check if the pattern repeats enough times
        if np.array_equal(pattern * repetitions, seq[: pattern_length * repetitions]):
            return (pattern, repetitions)

    return None, None


def _pattern_repr(pattern: Sequence[int]) -> str:
    """Turn pattern into a slice object if possible."""
    start = pattern[0]
    stop = pattern[-1] + 1
    if len(pattern) > 1:
        step = pattern[1] - pattern[0]
    else:
        step = 1
    if all(pattern[i] == pattern[0] + i * step for i in range(1, len(pattern))):
        if step == 1:
            if start == 0:
                return f"slice({stop})"
            return f"slice({start}, {stop})"
        return f"slice({start}, {stop}, {step})"
    return repr(pattern)


class _Repr:
    def __init__(self, string: str) -> None:
        self._string = string

    def __repr__(self) -> str:
        return self._string


def _expression_repr(expr: tuple[Sequence[int], Sequence[int]]) -> _Repr:
    """Try to represent an index expression as slice objects if possible."""
    e0, e1 = expr
    ptrn1, repeats = _find_pattern(e1)
    if ptrn1 is None:
        return _Repr(str(expr))
    ptrn0 = e0[:: len(ptrn1)]
    return _Repr(f"({_pattern_repr(ptrn0)}, {_pattern_repr(ptrn1)})")



from __future__ import annotations

import contextlib
import math
import warnings
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from annotated_types import Ge, Gt  # noqa: TCH002
from pydantic import Field, field_validator, model_validator

from useq._point_visiting import OrderMode, TraversalOrder
from useq._position import (
    AbsolutePosition,
    PositionT,
    RelativePosition,
    _MultiPointPlan,
)

if TYPE_CHECKING:
    from typing_extensions import Annotated, Self, TypeAlias

    PointGenerator: TypeAlias = Callable[
        [np.random.RandomState, int, float, float], Iterable[tuple[float, float]]
    ]

MIN_RANDOM_POINTS = 10000


class RelativeTo(Enum):
    """Where the coordinates of the grid are relative to.

    Attributes
    ----------
    center : Literal['center']
        Grid is centered around the origin.
    top_left : Literal['top_left']
        Grid is positioned such that the top left corner is at the origin.
    """

    center: str = "center"
    top_left: str = "top_left"


# used in iter_indices below, to determine the order in which indices are yielded
class _GridPlan(_MultiPointPlan[PositionT]):
    """Base class for all grid plans.

    Attributes
    ----------
    overlap : float | Tuple[float, float]
        Overlap between grid positions in percent. If a single value is provided, it is
        used for both x and y. If a tuple is provided, the first value is used
        for x and the second for y.
    mode : OrderMode
        Define the ways of ordering the grid positions. Options are
        row_wise, column_wise, row_wise_snake, column_wise_snake and spiral.
        By default, row_wise_snake.
    fov_width : Optional[float]
        Width of the field of view in microns.  If not provided, acquisition engines
        should use current width of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    fov_height : Optional[float]
        Height of the field of view in microns. If not provided, acquisition engines
        should use current height of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    """

    overlap: Tuple[float, float] = Field((0.0, 0.0), frozen=True)
    mode: OrderMode = Field(OrderMode.row_wise_snake, frozen=True)

    @field_validator("overlap", mode="before")
    def _validate_overlap(cls, v: Any) -> Tuple[float, float]:
        with contextlib.suppress(TypeError, ValueError):
            v = float(v)
        if isinstance(v, float):
            return (v,) * 2
        if isinstance(v, Sequence) and len(v) == 2:
            return float(v[0]), float(v[1])
        raise ValueError(  # pragma: no cover
            "overlap must be a float or a tuple of two floats"
        )

    def _offset_x(self, dx: float) -> float:
        raise NotImplementedError

    def _offset_y(self, dy: float) -> float:
        raise NotImplementedError

    def _nrows(self, dy: float) -> int:
        """Return the number of rows, given a grid step size."""
        raise NotImplementedError

    def _ncolumns(self, dx: float) -> int:
        """Return the number of columns, given a grid step size."""
        raise NotImplementedError

    def num_positions(self) -> int:
        """Return the number of individual positions in the grid.

        Note: For GridFromEdges and GridWidthHeight, this will depend on field of view
        size. If no field of view size is provided, the number of positions will be 1.
        """
        if isinstance(self, (GridFromEdges, GridWidthHeight)) and (
            # type ignore is because mypy thinks self is Never here...
            self.fov_width is None or self.fov_height is None  # type: ignore [attr-defined]
        ):
            raise ValueError(
                "Retrieving the number of positions in a GridFromEdges or "
                "GridWidthHeight plan requires the field of view size to be set."
            )

        dx, dy = self._step_size(self.fov_width or 1, self.fov_height or 1)
        rows = self._nrows(dy)
        cols = self._ncolumns(dx)
        return rows * cols

    def iter_grid_positions(
        self,
        fov_width: float | None = None,
        fov_height: float | None = None,
        *,
        order: OrderMode | None = None,
    ) -> Iterator[PositionT]:
        """Iterate over all grid positions, given a field of view size."""
        _fov_width = fov_width or self.fov_width or 1.0
        _fov_height = fov_height or self.fov_height or 1.0
        order = self.mode if order is None else OrderMode(order)

        dx, dy = self._step_size(_fov_width, _fov_height)
        rows = self._nrows(dy)
        cols = self._ncolumns(dx)
        x0 = self._offset_x(dx)
        y0 = self._offset_y(dy)

        pos_cls = RelativePosition if self.is_relative else AbsolutePosition
        for idx, (r, c) in enumerate(order.generate_indices(rows, cols)):
            yield pos_cls(
                x=x0 + c * dx,
                y=y0 - r * dy,
                row=r,
                col=c,
                name=f"{str(idx).zfill(4)}",
            )

    def __iter__(self) -> Iterator[PositionT]:  # type: ignore [override]
        yield from self.iter_grid_positions()

    def _step_size(self, fov_width: float, fov_height: float) -> Tuple[float, float]:
        dx = fov_width - (fov_width * self.overlap[0]) / 100
        dy = fov_height - (fov_height * self.overlap[1]) / 100
        return dx, dy


class GridFromEdges(_GridPlan[AbsolutePosition]):
    """Yield absolute stage positions to cover a bounded area.

    The bounded area is defined by top, left, bottom and right edges in
    stage coordinates.

    Attributes
    ----------
    top : float
        Top stage position of the bounding area
    left : float
        Left stage position of the bounding area
    bottom : float
        Bottom stage position of the bounding area
    right : float
        Right stage position of the bounding area
    overlap : float | Tuple[float, float]
        Overlap between grid positions in percent. If a single value is provided, it is
        used for both x and y. If a tuple is provided, the first value is used
        for x and the second for y.
    mode : OrderMode
        Define the ways of ordering the grid positions. Options are
        row_wise, column_wise, row_wise_snake, column_wise_snake and spiral.
        By default, row_wise_snake.
    fov_width : Optional[float]
        Width of the field of view in microns.  If not provided, acquisition engines
        should use current width of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    fov_height : Optional[float]
        Height of the field of view in microns. If not provided, acquisition engines
        should use current height of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    """

    # everything but fov_width and fov_height is immutable
    top: float = Field(..., frozen=True)
    left: float = Field(..., frozen=True)
    bottom: float = Field(..., frozen=True)
    right: float = Field(..., frozen=True)

    @property
    def is_relative(self) -> bool:
        return False

    def _nrows(self, dy: float) -> int:
        total_height = abs(self.top - self.bottom) + dy
        return math.ceil(total_height / dy)

    def _ncolumns(self, dx: float) -> int:
        total_width = abs(self.right - self.left) + dx
        return math.ceil(total_width / dx)

    def _offset_x(self, dx: float) -> float:
        return min(self.left, self.right)

    def _offset_y(self, dy: float) -> float:
        return max(self.top, self.bottom)


class GridRowsColumns(_GridPlan[RelativePosition]):
    """Grid plan based on number of rows and columns.

    Attributes
    ----------
    rows: int
        Number of rows.
    columns: int
        Number of columns.
    relative_to : RelativeTo
        Point in the grid to which the coordinates are relative. If "center", the grid
        is centered around the origin. If "top_left", the grid is positioned such that
        the top left corner is at the origin.
    overlap : float | Tuple[float, float]
        Overlap between grid positions in percent. If a single value is provided, it is
        used for both x and y. If a tuple is provided, the first value is used
        for x and the second for y.
    mode : OrderMode
        Define the ways of ordering the grid positions. Options are
        row_wise, column_wise, row_wise_snake, column_wise_snake and spiral.
        By default, row_wise_snake.
    fov_width : Optional[float]
        Width of the field of view in microns.  If not provided, acquisition engines
        should use current width of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    fov_height : Optional[float]
        Height of the field of view in microns. If not provided, acquisition engines
        should use current height of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    """

    # everything but fov_width and fov_height is immutable
    rows: int = Field(..., frozen=True, ge=1)
    columns: int = Field(..., frozen=True, ge=1)
    relative_to: RelativeTo = Field(RelativeTo.center, frozen=True)

    def _nrows(self, dy: float) -> int:
        return self.rows

    def _ncolumns(self, dx: float) -> int:
        return self.columns

    def _offset_x(self, dx: float) -> float:
        return (
            -((self.columns - 1) * dx) / 2
            if self.relative_to == RelativeTo.center
            else 0.0
        )

    def _offset_y(self, dy: float) -> float:
        return (
            ((self.rows - 1) * dy) / 2 if self.relative_to == RelativeTo.center else 0.0
        )


GridRelative = GridRowsColumns


class GridWidthHeight(_GridPlan[RelativePosition]):
    """Grid plan based on total width and height.

    Attributes
    ----------
    width: float
        Minimum total width of the grid, in microns. (may be larger based on fov_width)
    height: float
        Minimum total height of the grid, in microns. (may be larger based on
        fov_height)
    relative_to : RelativeTo
        Point in the grid to which the coordinates are relative. If "center", the grid
        is centered around the origin. If "top_left", the grid is positioned such that
        the top left corner is at the origin.
    overlap : float | Tuple[float, float]
        Overlap between grid positions in percent. If a single value is provided, it is
        used for both x and y. If a tuple is provided, the first value is used
        for x and the second for y.
    mode : OrderMode
        Define the ways of ordering the grid positions. Options are
        row_wise, column_wise, row_wise_snake, column_wise_snake and spiral.
        By default, row_wise_snake.
    fov_width : Optional[float]
        Width of the field of view in microns.  If not provided, acquisition engines
        should use current width of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    fov_height : Optional[float]
        Height of the field of view in microns. If not provided, acquisition engines
        should use current height of the FOV based on the current objective and camera.
        Engines MAY override this even if provided.
    """

    width: float = Field(..., frozen=True, gt=0)
    height: float = Field(..., frozen=True, gt=0)
    relative_to: RelativeTo = Field(RelativeTo.center, frozen=True)

    def _nrows(self, dy: float) -> int:
        return math.ceil(self.height / dy)

    def _ncolumns(self, dx: float) -> int:
        return math.ceil(self.width / dx)

    def _offset_x(self, dx: float) -> float:
        return (
            -((self._ncolumns(dx) - 1) * dx) / 2
            if self.relative_to == RelativeTo.center
            else 0.0
        )

    def _offset_y(self, dy: float) -> float:
        return (
            ((self._nrows(dy) - 1) * dy) / 2
            if self.relative_to == RelativeTo.center
            else 0.0
        )


# ------------------------ RANDOM ------------------------


class Shape(Enum):
    """Shape of the bounding box for random points.

    Attributes
    ----------
    ELLIPSE : Literal['ellipse']
        The bounding box is an ellipse.
    RECTANGLE : Literal['rectangle']
        The bounding box is a rectangle.
    """

    ELLIPSE = "ellipse"
    RECTANGLE = "rectangle"


class RandomPoints(_MultiPointPlan[RelativePosition]):
    """Yield random points in a specified geometric shape.

    Attributes
    ----------
    num_points : int
        Number of points to generate.
    max_width : float
        Maximum width of the bounding box in microns.
    max_height : float
        Maximum height of the bounding box in microns.
    shape : Shape
        Shape of the bounding box. Current options are "ellipse" and "rectangle".
    random_seed : Optional[int]
        Random numpy seed that should be used to generate the points. If None, a random
        seed will be used.
    allow_overlap : bool
        By defaut, True. If False and `fov_width` and `fov_height` are specified, points
        will not overlap and will be at least `fov_width` and `fov_height apart.
    order : TraversalOrder
        Order in which the points will be visited. If None, order is simply the order
        in which the points are generated (random).  Use 'nearest_neighbor' or
        'two_opt' to order the points in a more structured way.
    start_at : int | RelativePosition
        Position or index of the point to start at. This is only used if `order` is
        'nearest_neighbor' or 'two_opt'.  If a position is provided, it will *always*
        be included in the list of points. If an index is provided, it must be less than
        the number of points, and corresponds to the index of the (randomly generated)
        points; this likely only makes sense when `random_seed` is provided.
    """

    num_points: Annotated[int, Gt(0)]
    max_width: Annotated[float, Gt(0)] = 1
    max_height: Annotated[float, Gt(0)] = 1
    shape: Shape = Shape.ELLIPSE
    random_seed: Optional[int] = None
    allow_overlap: bool = True
    order: Optional[TraversalOrder] = TraversalOrder.TWO_OPT
    start_at: Union[RelativePosition, Annotated[int, Ge(0)]] = 0

    @model_validator(mode="after")
    def _validate_startat(self) -> Self:
        if isinstance(self.start_at, int) and self.start_at > (self.num_points - 1):
            warnings.warn(
                "start_at is greater than the number of points. "
                "Setting start_at to last point.",
                stacklevel=2,
            )
            self.start_at = self.num_points - 1
        return self

    def __iter__(self) -> Iterator[RelativePosition]:  # type: ignore [override]
        seed = np.random.RandomState(self.random_seed)
        func = _POINTS_GENERATORS[self.shape]

        points: list[Tuple[float, float]] = []
        needed_points = self.num_points
        start_at = self.start_at
        if isinstance(start_at, RelativePosition):
            points = [(start_at.x, start_at.y)]
            needed_points -= 1
            start_at = 0

        # in the easy case, just generate the requested number of points
        if self.allow_overlap or self.fov_width is None or self.fov_height is None:
            _points = func(seed, needed_points, self.max_width, self.max_height)
            points.extend(_points)

        else:
            # if we need to avoid overlap, generate points, check if they are valid, and
            # repeat until we have enough
            per_iter = needed_points
            tries = 0
            while tries < MIN_RANDOM_POINTS and len(points) < self.num_points:
                candidates = func(seed, per_iter, self.max_width, self.max_height)
                tries += per_iter
                for p in candidates:
                    if _is_a_valid_point(points, *p, self.fov_width, self.fov_height):
                        points.append(p)
                        if len(points) >= self.num_points:
                            break

            if len(points) < self.num_points:
                warnings.warn(
                    f"Unable to generate {self.num_points} non-overlapping points. "
                    f"Only {len(points)} points were found.",
                    stacklevel=2,
                )

        if self.order is not None:
            points = self.order(points, start_at=start_at)  # type: ignore [assignment]

        for idx, (x, y) in enumerate(points):
            yield RelativePosition(x=x, y=y, name=f"{str(idx).zfill(4)}")

    def num_positions(self) -> int:
        return self.num_points


def _is_a_valid_point(
    points: list[Tuple[float, float]],
    x: float,
    y: float,
    min_dist_x: float,
    min_dist_y: float,
) -> bool:
    """Return True if the the point is at least min_dist away from all the others.

    note: using Manhattan distance.
    """
    return not any(
        abs(x - point_x) < min_dist_x and abs(y - point_y) < min_dist_y
        for point_x, point_y in points
    )


def _random_points_in_ellipse(
    seed: np.random.RandomState, n_points: int, max_width: float, max_height: float
) -> np.ndarray:
    """Generate a random point around a circle with center (0, 0).

    The point is within +/- radius_x and +/- radius_y at a random angle.
    """
    points = seed.uniform(0, 1, size=(n_points, 3))
    xy = points[:, :2]
    angle = points[:, 2] * 2 * np.pi
    xy[:, 0] *= (max_width / 2) * np.cos(angle)
    xy[:, 1] *= (max_height / 2) * np.sin(angle)
    return xy


def _random_points_in_rectangle(
    seed: np.random.RandomState, n_points: int, max_width: float, max_height: float
) -> np.ndarray:
    """Generate a random point around a rectangle with center (0, 0).

    The point is within the bounding box (-width/2, -height/2, width, height).
    """
    xy = seed.uniform(0, 1, size=(n_points, 2))
    xy[:, 0] = (xy[:, 0] * max_width) - (max_width / 2)
    xy[:, 1] = (xy[:, 1] * max_height) - (max_height / 2)
    return xy


_POINTS_GENERATORS: dict[Shape, PointGenerator] = {
    Shape.ELLIPSE: _random_points_in_ellipse,
    Shape.RECTANGLE: _random_points_in_rectangle,
}


# all of these support __iter__() -> Iterator[PositionBase] and num_positions() -> int
RelativeMultiPointPlan = Union[
    GridRowsColumns, GridWidthHeight, RandomPoints, RelativePosition
]
AbsoluteMultiPointPlan = Union[GridFromEdges]
MultiPointPlan = Union[AbsoluteMultiPointPlan, RelativeMultiPointPlan]



from __future__ import annotations

from enum import Enum
from functools import partial
from typing import Callable, Iterable, Iterator, Tuple

import numpy as np

# ----------------------------- Random Points -----------------------------------


class OrderMode(Enum):
    """Order in which grid positions will be iterated.

    Attributes
    ----------
    row_wise : Literal['row_wise']
        Iterate row by row.
    column_wise : Literal['column_wise']
        Iterate column by column.
    row_wise_snake : Literal['row_wise_snake']
        Iterate row by row, but alternate the direction of the columns.
    column_wise_snake : Literal['column_wise_snake']
        Iterate column by column, but alternate the direction of the rows.
    spiral : Literal['spiral']
        Iterate in a spiral pattern, starting from the center.
    """

    row_wise = "row_wise"
    column_wise = "column_wise"
    row_wise_snake = "row_wise_snake"
    column_wise_snake = "column_wise_snake"
    spiral = "spiral"

    def generate_indices(self, rows: int, columns: int) -> Iterator[Tuple[int, int]]:
        """Generate indices for the given grid size."""
        return _INDEX_GENERATORS[self](rows, columns)


def _spiral_indices(
    rows: int, columns: int, center_origin: bool = False
) -> Iterator[Tuple[int, int]]:
    """Return a spiral iterator over a 2D grid.

    Parameters
    ----------
    rows : int
        Number of rows.
    columns : int
        Number of columns.
    center_origin : bool
        If center_origin is True, all indices are centered around (0, 0), and some will
        be negative. Otherwise, the indices are centered around (rows//2, columns//2)

    Yields
    ------
    (x, y) : tuple[int, int]
        Indices of the next element in the spiral.
    """
    # direction: first down and then clockwise (assuming positive Y is down)

    x = y = 0
    if center_origin:  # see docstring
        xshift = yshift = 0
    else:
        xshift = (columns - 1) // 2
        yshift = (rows - 1) // 2
    dx = 0
    dy = -1
    for _ in range(max(columns, rows) ** 2):
        if (-columns / 2 < x <= columns / 2) and (-rows / 2 < y <= rows / 2):
            yield y + yshift, x + xshift
        if x == y or (x < 0 and x == -y) or (x > 0 and x == 1 - y):
            dx, dy = -dy, dx
        x, y = x + dx, y + dy


# function that iterates indices (row, col) in a grid where (0, 0) is the top left
def _rect_indices(
    rows: int, columns: int, snake: bool = False, row_wise: bool = True
) -> Iterator[Tuple[int, int]]:
    """Return a row or column-wise iterator over a 2D grid."""
    c, r = np.meshgrid(np.arange(columns), np.arange(rows))
    if snake:
        if row_wise:
            c[1::2, :] = c[1::2, :][:, ::-1]
        else:
            r[:, 1::2] = r[:, 1::2][::-1, :]
    return zip(r.ravel(), c.ravel()) if row_wise else zip(r.T.ravel(), c.T.ravel())


IndexGenerator = Callable[[int, int], Iterator[Tuple[int, int]]]
_INDEX_GENERATORS: dict[OrderMode, IndexGenerator] = {
    OrderMode.row_wise: partial(_rect_indices, snake=False, row_wise=True),
    OrderMode.column_wise: partial(_rect_indices, snake=False, row_wise=False),
    OrderMode.row_wise_snake: partial(_rect_indices, snake=True, row_wise=True),
    OrderMode.column_wise_snake: partial(_rect_indices, snake=True, row_wise=False),
    OrderMode.spiral: _spiral_indices,
}

# ----------------------------- Random Points -----------------------------------


class TraversalOrder(Enum):
    NEAREST_NEIGHBOR = "nearest_neighbor"
    TWO_OPT = "two_opt"
    RANDOM = "random"

    def order_points(self, points: np.ndarray, start_at: int = 0) -> np.ndarray:
        """Return the order of points based on the traversal order."""
        if len(points) <= 2:  # no sense in optimizing
            return np.arange(len(points))
        start_at = min(start_at, len(points) - 1)
        if self == TraversalOrder.NEAREST_NEIGHBOR:
            return _nearest_neighbor_order(points, start_at)
        if self == TraversalOrder.TWO_OPT:
            return _two_opt_order(points, start_at)
        if self == TraversalOrder.RANDOM:
            return np.random.permutation(len(points))
        raise ValueError(f"Unknown traversal order: {self}")  # pragma: no cover

    def __call__(
        self, points: Iterable[tuple[float, float]], start_at: int = 0
    ) -> np.ndarray:
        """Sort the points based on the traversal order."""
        points = np.asarray(points)
        order = self.order_points(points, start_at=start_at)
        return points[order]  # type: ignore [no-any-return]


def _nearest_neighbor_order(points: np.ndarray, start_at: int = 0) -> np.ndarray:
    """Return the order of points based on the nearest neighbor algorithm.

    Parameters.
    ----------
    points : np.ndarray
        Array of 2D (Y, X) points in the format (n, 2).
    start_at : int, optional
        Index of the point to start at.  By default, the first point is used.

    Examples
    --------
    >>> points = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
    >>> order = _nearest_neighbor_order(points)
    >>> sorted = points[order]
    """
    n = len(points)
    visited = np.zeros(n, dtype=bool)
    order = np.zeros(n, dtype=int)
    order[0] = start_at
    visited[start_at] = True

    LARGE_NUMBER = 1e9
    # NOTE: at ~500+ points, scipy.spatial.cKDTree would begin to be faster
    # but it's a new dependency and may not be a common use case
    for i in range(1, n):
        # calculate the distance from the last visited point to all other points
        dist = np.linalg.norm(points - points[order[i - 1]], axis=1)
        # find the nearest point that has not been visited
        next_point = np.argmin(dist + visited * LARGE_NUMBER)
        # store it and mark it as visited
        order[i] = next_point
        visited[next_point] = True

    return order


def _two_opt_order(
    points: np.ndarray, start_at: int = 0, improvement_threshold: float = 0.05
) -> np.ndarray:
    """Return the order of points based on the 2-opt algorithm.

    https://en.wikipedia.org/wiki/2-opt

    Parameters
    ----------
    points : np.ndarray
        Array of 2D (Y, X) points in the format (n, 2).
    start_at : int, optional
        Index of the point to start at. By default, the first point is used.
    improvement_threshold : float, optional
        The minimum improvement factor required to continue the optimization.
        By default, 0.05.

    Examples
    --------
    >>> points = np.random.rand(100, 2)
    >>> order = _two_opt_order(points)
    >>> sorted = points[order]
    """
    n = points.shape[0]
    route = np.arange(n)

    if start_at != 0:
        route = np.roll(route, -start_at)

    dist_matrix = _distance_matrix(points)

    # this will track the best distance found so far
    best_distance = _total_distance(points[route])
    improvement_factor = 1.0
    while improvement_factor > improvement_threshold:
        distance_to_beat = best_distance
        for i in range(1, n - 2):
            for k in range(i + 1, n):
                # Calculate the distances involved in the potential swap
                ri = route[i]
                ri1 = route[i - 1]
                rk = route[k]
                y = route[(k + 1) % n]

                dist_before = dist_matrix[ri1, ri] + dist_matrix[rk, y]
                dist_after = dist_matrix[ri1, rk] + dist_matrix[ri, y]

                # If the new distance is better, perform the swap
                if dist_after < dist_before:
                    # Reverse the order of all elements from element i to element k.
                    route[i : k + 1] = route[i : k + 1][::-1]
                    best_distance = best_distance - dist_before + dist_after

        improvement_factor = 1 - best_distance / distance_to_beat

    return route


def _total_distance(points: np.ndarray) -> float:
    # Calculate the total Euclidean distance of the route traversing the given points
    # in the order provided
    diffs = points - np.roll(points, shift=1, axis=0)
    return np.sum(np.linalg.norm(diffs, axis=1))  # type: ignore [no-any-return]


def _distance_matrix(points: np.ndarray) -> np.ndarray:
    # Calculate the distance matrix (euclidean distance between each pair of points)
    return np.sqrt(np.sum((points[:, None] - points) ** 2, axis=2))  # type: ignore [no-any-return]



import json
from pathlib import Path

import pytest

from useq import MDASequence


@pytest.mark.parametrize("ext", ["json", "yaml"])
def test_serialization(mda1: MDASequence, ext: str) -> None:
    FILE = Path(__file__).parent / "fixtures" / f"mda.{ext}"
    text = FILE.read_text()
    mda = MDASequence.from_file(str(FILE))
    assert mda == mda1
    if ext == "json":
        assert json.loads(mda.model_dump_json(exclude={"uid"})) == json.loads(text)
    else:
        assert mda.yaml() == text

    it = iter(mda)
    for _ in range(20):
        if ext == "json":
            assert next(it).model_dump_json()
        else:
            assert next(it).yaml()



from __future__ import annotations

import re
from datetime import timedelta
from typing import TYPE_CHECKING, Literal, NamedTuple, TypeVar

if TYPE_CHECKING:
    from typing import Final

    from typing_extensions import TypeGuard

    import useq
    from useq._time import SinglePhaseTimePlan

    KT = TypeVar("KT")
    VT = TypeVar("VT")


# could be an enum, but this more easily allows Axis.Z to be a string
class Axis:
    """Recognized axis names."""

    TIME: Final[Literal["t"]] = "t"
    POSITION: Final[Literal["p"]] = "p"
    GRID: Final[Literal["g"]] = "g"
    CHANNEL: Final[Literal["c"]] = "c"
    Z: Final[Literal["z"]] = "z"


# note: order affects the default axis_order in MDASequence
AXES: Final[tuple[str, ...]] = (
    Axis.TIME,
    Axis.POSITION,
    Axis.GRID,
    Axis.CHANNEL,
    Axis.Z,
)


class TimeEstimate(NamedTuple):
    """Record of time estimation results.

    Attributes
    ----------
    total_duration : float
        Estimated total duration of the experiment, in seconds.
    per_t_duration : float
        Estimated duration of a single timepoint, in seconds.
    time_interval_exceeded : bool
        Whether the time interval between timepoints is shorter than the time it takes
        to acquire the data.  In a multi-phase time plan, this is True if any of the
        phases have this property.
    """

    total_duration: float
    per_t_duration: float
    time_interval_exceeded: bool

    def __add__(self, other: object) -> TimeEstimate:
        """Add two TimeEstimates."""
        if not isinstance(other, TimeEstimate):
            return NotImplemented  # pragma: no cover
        return TimeEstimate(
            self.total_duration + other.total_duration,
            self.per_t_duration + other.per_t_duration,
            self.time_interval_exceeded or other.time_interval_exceeded,
        )


def estimate_sequence_duration(seq: useq.MDASequence) -> TimeEstimate:
    """Estimate the duration of an MDASequence.

    Notable mis-estimations may include:
    - when the time interval is shorter than the time it takes to acquire the data
      and any of the channels have `acquire_every` > 1
    - when channel exposure times are omitted. In this case, we assume 1ms exposure.

    Returns
    -------
    TimeEstimate
        A named 3-tuple with the following fields:
        - total_duration: float
            Estimated total duration of the experiment, in seconds.
        - per_t_duration: float
            Estimated duration of a single timepoint, in seconds.
        - time_interval_exceeded: bool
            Whether the time interval between timepoints is shorter than the time it
            takes to acquire the data
    """
    stage_positions = tuple(seq.stage_positions)
    if not any(_has_axes(p.sequence) for p in stage_positions):
        # the simple case: no axes to iterate over in any of the positions
        return _estimate_simple_sequence_duration(seq)

    estimate = TimeEstimate(0.0, 0.0, False)
    parent_seq = seq.replace(stage_positions=[])
    for p in stage_positions:
        if not _has_axes(p.sequence):
            sub_seq = parent_seq
        else:
            updates = {
                field: getattr(parent_seq, field)
                for field in ("time_plan", "z_plan", "grid_plan", "channels")
                if getattr(parent_seq, field) and not getattr(p.sequence, field)
            }
            sub_seq = p.sequence.model_copy(update=updates)
        estimate += _estimate_simple_sequence_duration(sub_seq)
    return estimate


def _estimate_simple_sequence_duration(seq: useq.MDASequence) -> TimeEstimate:
    """Estimate the duration of an MDASequence with no sub-pos axes to iterate over.

    Helper function for estimate_sequence_duration.
    """
    n_positions = len(seq.stage_positions or [0])
    num_z = seq.z_plan.num_positions() if seq.z_plan else 1
    num_grid = seq.grid_plan.num_positions() if seq.grid_plan else 1

    # NOTE:
    # technically some channels might have ch.aquire_every > 1
    # we don't account for that here.  The ONLY case where this will lead to erroneous
    # estimates is when the time plan interval is shorter than the time it takes to
    # acquire the data (i.e. when `time_interval_exceeded` is True`).  In that case,
    # we will overestimate the total time a little bit.
    s_per_timepoint: float = 0.0
    for ch in seq.channels:
        # note, using 1ms instead of 0ms when channel exposure has been omitted
        # this will likely underestimate the real time (which will then be determined
        # by the "current" exposure time), but it's better than 0 or a higher magic num.
        exposure_s = (ch.exposure or 1) / 1000
        num_ch_z = num_z if ch.do_stack else 1
        s_per_timepoint += exposure_s * n_positions * num_ch_z * num_grid

    t_interval_exceeded = False
    if tplan := seq.time_plan:
        phases = tplan.phases if hasattr(tplan, "phases") else [tplan]
        tot_duration = 0.0
        for phase in phases:
            phase_duration, exceeded = _time_phase_duration(phase, s_per_timepoint)
            tot_duration += phase_duration
            t_interval_exceeded = t_interval_exceeded or exceeded
    else:
        tot_duration = s_per_timepoint

    return TimeEstimate(tot_duration, s_per_timepoint, t_interval_exceeded)


def _time_phase_duration(
    phase: SinglePhaseTimePlan, s_per_timepoint: float
) -> tuple[float, bool]:
    """Calculate duration for a single time plan phase."""
    time_interval_s = phase.interval.total_seconds()

    if time_interval_exceeded := (s_per_timepoint > time_interval_s):
        # the real interval between timepoints will be the greater of the
        # prescribed interval (according to the time plan) and the time it takes
        # to actually acquire the data
        time_interval_s = s_per_timepoint

    tot_duration = (phase.num_timepoints() - 1) * time_interval_s + s_per_timepoint
    return tot_duration, time_interval_exceeded


def _has_axes(seq: useq.MDASequence | None) -> TypeGuard[useq.MDASequence]:
    """Return True if the sequence has anything to iterate over."""
    if seq is None:
        return False
    return bool(
        seq.time_plan is not None
        or seq.stage_positions
        or seq.z_plan is not None
        or seq.channels
        or seq.grid_plan is not None
    )


# vendored from pydantic v1
def parse_duration(value: str | bytes | int | float) -> timedelta:  # pragma: no cover
    """
    Parse a duration int/float/string and return a datetime.timedelta.

    The preferred format for durations in Django is '%d %H:%M:%S.%f'.

    Also supports ISO 8601 representation.
    """
    if isinstance(value, timedelta):
        return value

    if isinstance(value, (int, float)):
        # below code requires a string
        value = f"{value:f}"
    elif isinstance(value, bytes):
        value = value.decode()

    try:
        match = standard_duration_re.match(value) or iso8601_duration_re.match(value)
    except TypeError:
        raise TypeError(
            "invalid type; expected timedelta, string, bytes, int or float"
        ) from None

    if not match:
        raise ValueError(f"{value!r} is not a valid duration")

    kw = match.groupdict()
    sign = -1 if kw.pop("sign", "+") == "-" else 1
    if kw.get("microseconds"):
        kw["microseconds"] = kw["microseconds"].ljust(6, "0")

    if kw.get("seconds") and kw.get("microseconds") and kw["seconds"].startswith("-"):
        kw["microseconds"] = "-" + kw["microseconds"]

    kw_ = {k: float(v) for k, v in kw.items() if v is not None}

    return sign * timedelta(**kw_)


standard_duration_re = re.compile(
    r"^"
    r"(?:(?P<days>-?\d+) (days?, )?)?"
    r"((?:(?P<hours>-?\d+):)(?=\d+:\d+))?"
    r"(?:(?P<minutes>-?\d+):)?"
    r"(?P<seconds>-?\d+)"
    r"(?:\.(?P<microseconds>\d{1,6})\d{0,6})?"
    r"$"
)

# Support the sections of ISO 8601 date representation that are accepted by timedelta
iso8601_duration_re = re.compile(
    r"^(?P<sign>[-+]?)"
    r"P"
    r"(?:(?P<days>\d+(.\d+)?)D)?"
    r"(?:T"
    r"(?:(?P<hours>\d+(.\d+)?)H)?"
    r"(?:(?P<minutes>\d+(.\d+)?)M)?"
    r"(?:(?P<seconds>\d+(.\d+)?)S)?"
    r")?"
    r"$"
)



from typing import TYPE_CHECKING, Generic, Iterator, Optional, SupportsIndex, TypeVar

from pydantic import Field

from useq._base_model import FrozenModel, MutableModel

if TYPE_CHECKING:
    from typing_extensions import Self

    from useq import MDASequence


class PositionBase(MutableModel):
    """Define a position in 3D space.

    Any of the attributes can be `None` to indicate that the position is not
    defined.  For engines implementing support for useq, a position of `None` implies
    "do not move" or "stay at current position" on that axis.

    Attributes
    ----------
    x : float | None
        X position in microns.
    y : float | None
        Y position in microns.
    z : float | None
        Z position in microns.
    name : str | None
        Optional name for the position.
    sequence : MDASequence | None
        Optional MDASequence relative this position.
    row : int | None
        Optional row index, when used in a grid.
    col : int | None
        Optional column index, when used in a grid.
    """

    x: Optional[float] = None
    y: Optional[float] = None
    z: Optional[float] = None
    name: Optional[str] = None
    sequence: Optional["MDASequence"] = None

    # excluded from serialization
    row: Optional[int] = Field(default=None, exclude=True)
    col: Optional[int] = Field(default=None, exclude=True)

    def __add__(self, other: "RelativePosition") -> "Self":
        """Add two positions together to create a new position."""
        if not isinstance(other, RelativePosition):  # pragma: no cover
            return NotImplemented
        if (x := self.x) is not None and other.x is not None:
            x += other.x
        if (y := self.y) is not None and other.y is not None:
            y += other.y
        if (z := self.z) is not None and other.z is not None:
            z += other.z
        if (name := self.name) and other.name:
            name = f"{name}_{other.name}"
        kwargs = {**self.model_dump(), "x": x, "y": y, "z": z, "name": name}
        return type(self).model_construct(**kwargs)  # type: ignore [return-value]

    def __round__(self, ndigits: "SupportsIndex | None" = None) -> "Self":
        """Round the position to the given number of decimal places."""
        kwargs = {
            **self.model_dump(),
            "x": round(self.x, ndigits) if self.x is not None else None,
            "y": round(self.y, ndigits) if self.y is not None else None,
            "z": round(self.z, ndigits) if self.z is not None else None,
        }
        # not sure why these Self types are not working
        return type(self).model_construct(**kwargs)  # type: ignore [return-value]


class AbsolutePosition(PositionBase, FrozenModel):
    """An absolute position in 3D space."""

    @property
    def is_relative(self) -> bool:
        return False


Position = AbsolutePosition  # for backwards compatibility
PositionT = TypeVar("PositionT", bound=PositionBase)


class _MultiPointPlan(MutableModel, Generic[PositionT]):
    """Any plan that yields multiple positions."""

    fov_width: Optional[float] = None
    fov_height: Optional[float] = None

    @property
    def is_relative(self) -> bool:
        return True

    def __iter__(self) -> Iterator[PositionT]:  # type: ignore [override]
        raise NotImplementedError("This method must be implemented by subclasses.")

    def num_positions(self) -> int:
        raise NotImplementedError("This method must be implemented by subclasses.")

    def plot(self) -> None:
        """Plot the positions in the plan."""
        from useq._plot import plot_points

        plot_points(self)


class RelativePosition(PositionBase, _MultiPointPlan["RelativePosition"]):
    """A relative position in 3D space.

    Relative positions also support `fov_width` and `fov_height` attributes, and can
    be used to define a single field of view for a "multi-point" plan.
    """

    x: float = 0
    y: float = 0
    z: float = 0

    def __iter__(self) -> Iterator["RelativePosition"]:  # type: ignore [override]
        yield self

    def num_positions(self) -> int:
        return 1



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Optional, get_args

import pytest

from useq import (
    GridFromEdges,
    GridRowsColumns,
    GridWidthHeight,
    RandomPoints,
    RelativeMultiPointPlan,
    RelativePosition,
    TraversalOrder,
)
from useq._point_visiting import OrderMode, _rect_indices, _spiral_indices

if TYPE_CHECKING:
    from useq._position import PositionBase

EXPECT = {
    (True, False): [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)],
    (True, True): [(0, 0), (0, 1), (1, 1), (1, 0), (2, 0), (2, 1)],
    (False, False): [(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1)],
    (False, True): [(0, 0), (1, 0), (2, 0), (2, 1), (1, 1), (0, 1)],
}


@pytest.mark.parametrize("row_wise", [True, False], ids=["row_wise", "col_wise"])
@pytest.mark.parametrize("snake", [True, False], ids=["snake", "normal"])
def test_grid_indices(row_wise: bool, snake: bool) -> None:
    indices = _rect_indices(3, 2, snake=snake, row_wise=row_wise)
    assert list(indices) == EXPECT[(row_wise, snake)]


def test_spiral_indices() -> None:
    assert list(_spiral_indices(2, 3)) == [
        (0, 1),
        (0, 2),
        (1, 2),
        (1, 1),
        (1, 0),
        (0, 0),
    ]
    assert list(_spiral_indices(2, 3, center_origin=True)) == [
        (0, 0),
        (0, 1),
        (1, 1),
        (1, 0),
        (1, -1),
        (0, -1),
    ]


def test_position_equality():
    """Order of grid positions should only change the order in which they are yielded"""

    def positions_without_name(
        positions: Iterable[PositionBase],
    ) -> set[tuple[float, float, bool]]:
        """Create a set of tuples of GridPosition attributes excluding 'name'"""
        return {(pos.x, pos.y, pos.is_relative) for pos in positions}

    t1 = GridRowsColumns(rows=3, columns=3, mode=OrderMode.spiral)
    spiral_pos = positions_without_name(t1.iter_grid_positions(1, 1))

    t2 = GridRowsColumns(rows=3, columns=3, mode=OrderMode.row_wise)
    row_pos = positions_without_name(t2.iter_grid_positions(1, 1))

    t3 = GridRowsColumns(rows=3, columns=3, mode="row_wise_snake")
    snake_row_pos = positions_without_name(t3.iter_grid_positions(1, 1))

    t4 = GridRowsColumns(rows=3, columns=3, mode=OrderMode.column_wise)
    col_pos = positions_without_name(t4.iter_grid_positions(1, 1))

    t5 = GridRowsColumns(rows=3, columns=3, mode=OrderMode.column_wise_snake)
    snake_col_pos = positions_without_name(t5.iter_grid_positions(1, 1))

    assert spiral_pos == row_pos == snake_row_pos == col_pos == snake_col_pos


def test_grid_type():
    g1 = GridRowsColumns(rows=2, columns=3)
    assert [(g.x, g.y) for g in g1.iter_grid_positions(1, 1)] == [
        (-1.0, 0.5),
        (0.0, 0.5),
        (1.0, 0.5),
        (1.0, -0.5),
        (0.0, -0.5),
        (-1.0, -0.5),
    ]
    assert g1.is_relative
    g2 = GridWidthHeight(width=3, height=2, fov_height=1, fov_width=1)
    assert [(g.x, g.y) for g in g2.iter_grid_positions()] == [
        (-1.0, 0.5),
        (0.0, 0.5),
        (1.0, 0.5),
        (1.0, -0.5),
        (0.0, -0.5),
        (-1.0, -0.5),
    ]
    assert g2.is_relative
    g3 = GridFromEdges(top=1, left=-1, bottom=-1, right=2)
    assert [(g.x, g.y) for g in g3.iter_grid_positions(1, 1)] == [
        (-1.0, 1.0),
        (0.0, 1.0),
        (1.0, 1.0),
        (2.0, 1.0),
        (2.0, 0.0),
        (1.0, 0.0),
        (0.0, 0.0),
        (-1.0, 0.0),
        (-1.0, -1.0),
        (0.0, -1.0),
        (1.0, -1.0),
        (2.0, -1.0),
    ]
    assert not g3.is_relative


def test_num_position_error() -> None:
    with pytest.raises(ValueError, match="plan requires the field of view size"):
        GridFromEdges(top=1, left=-1, bottom=-1, right=2).num_positions()

    with pytest.raises(ValueError, match="plan requires the field of view size"):
        GridWidthHeight(width=2, height=2).num_positions()


expected_rectangle = [(0.2, 1.1), (0.4, 0.2), (-0.3, 0.7)]
expected_ellipse = [(-0.9, -1.1), (0.9, -0.5), (-0.8, -0.4)]


@pytest.mark.parametrize("n_points", [3, 100])
@pytest.mark.parametrize("shape", ["rectangle", "ellipse"])
@pytest.mark.parametrize("seed", [None, 0])
def test_random_points(n_points: int, shape: str, seed: Optional[int]) -> None:
    rp = RandomPoints(
        num_points=n_points,
        max_width=4,
        max_height=5,
        shape=shape,
        random_seed=seed,
        allow_overlap=False,
        fov_width=0.5,
        fov_height=0.5,
    )

    if n_points == 3:
        expected = expected_rectangle if shape == "rectangle" else expected_ellipse
        values = [(round(g.x, 1), round(g.y, 1)) for g in rp]
        if seed is None:
            assert values != expected
        else:
            assert values == expected
    else:
        with pytest.raises(UserWarning, match="Unable to generate"):
            list(rp)


@pytest.mark.parametrize("order", list(TraversalOrder))
def test_traversal(order: TraversalOrder):
    pp = RandomPoints(
        num_points=30,
        max_height=3000,
        max_width=3000,
        order=order,
        random_seed=1,
        start_at=10,
        fov_height=300,
        fov_width=300,
        allow_overlap=False,
    )
    list(pp)


fov = {"fov_height": 200, "fov_width": 200}


@pytest.mark.parametrize(
    "obj",
    [
        GridRowsColumns(rows=1, columns=2, **fov),
        GridWidthHeight(width=10, height=10, **fov),
        RandomPoints(num_points=10, **fov),
        RelativePosition(**fov),
    ],
)
def test_points_plans(obj: RelativeMultiPointPlan, monkeypatch: pytest.MonkeyPatch):
    mpl = pytest.importorskip("matplotlib.pyplot")
    monkeypatch.setattr(mpl, "show", lambda: None)

    assert isinstance(obj, get_args(RelativeMultiPointPlan))
    assert all(isinstance(x, RelativePosition) for x in obj)
    assert isinstance(obj.num_positions(), int)

    obj.plot()



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from typing import Iterable, Mapping, Required, TypeAlias, TypedDict

    from useq._plate import WellPlate

    class KnownPlateKwargs(TypedDict, total=False):
        rows: Required[int]
        columns: Required[int]
        well_spacing: Required[tuple[float, float] | float]
        well_size: tuple[float, float] | float | None
        circular_wells: bool
        name: str

    PlateOrKwargs: TypeAlias = KnownPlateKwargs | WellPlate


_PLATE_REGISTRY: dict[str, PlateOrKwargs] = {
    "6-well": {"rows": 2, "columns": 3, "well_spacing": 39.12, "well_size": 34.8},
    "12-well": {"rows": 3, "columns": 4, "well_spacing": 26, "well_size": 22},
    "24-well": {"rows": 4, "columns": 6, "well_spacing": 19, "well_size": 15.6},
    "48-well": {"rows": 6, "columns": 8, "well_spacing": 13, "well_size": 11.1},
    "96-well": {"rows": 8, "columns": 12, "well_spacing": 9, "well_size": 6.4},
    "384-well": {
        "rows": 16,
        "columns": 24,
        "well_spacing": 4.5,
        "well_size": 3.4,
        "circular_wells": False,
    },
    "1536-well": {
        "rows": 32,
        "columns": 48,
        "well_spacing": 2.25,
        "well_size": 1.7,
        "circular_wells": False,
    },
    "coverslip-18mm-square": {
        "rows": 1,
        "columns": 1,
        "well_spacing": 0.0,
        "well_size": 18.0,
        "circular_wells": False,
        "name": "18mm coverslip",
    },
    "coverslip-22mm-square": {
        "rows": 1,
        "columns": 1,
        "well_spacing": 0.0,
        "well_size": 22.0,
        "circular_wells": False,
        "name": "22mm coverslip",
    },
}


@overload
def register_well_plates(
    plates: Mapping[str, PlateOrKwargs],
    /,
    **kwargs: PlateOrKwargs,
) -> None: ...
@overload
def register_well_plates(
    plates: Iterable[tuple[str, PlateOrKwargs]],
    /,
    **kwargs: PlateOrKwargs,
) -> None: ...
@overload
def register_well_plates(**kwargs: PlateOrKwargs) -> None: ...
def register_well_plates(
    plates: Mapping[str, PlateOrKwargs] | Iterable[tuple[str, PlateOrKwargs]] = (),
    /,
    **kwargs: PlateOrKwargs,
) -> None:
    """Register well-plate definitions to allow lookup by key.

    Added keys will override existing keys if they already exist.

    Values may either be WellPlate instances, or dictionaries with the following keys:
        - rows: Required[int]
        - columns: Required[int]
        - well_spacing: Required[tuple[float, float]]
        - well_size: tuple[float, float] | None
        - circular_wells: bool
        - name: str

    Examples
    --------
    >>> import useq
    >>> useq.register_well_plates(
    ...     {
    ...         "custom-square-plate": {
    ...             "rows": 8, "columns": 8, "well_spacing": 9.3, "well_size": 7.1
    ...         },
    ...         "silly-plate": {"rows": 1, "columns": 1, "well_spacing": 100}
    ...     }
    ... )
    """
    _PLATE_REGISTRY.update(plates, **kwargs)


def registered_well_plate_keys() -> set[str]:
    """Return a set of all registered well-plate keys.

    These keys may be used as an argument to `WellPlatePlan.plate` to select a plate
    definition.
    """
    return set(_PLATE_REGISTRY)



"""Implementation agnostic schema for multi-dimensional microscopy experiments."""

import warnings
from typing import Any

from useq._actions import AcquireImage, Action, HardwareAutofocus
from useq._channel import Channel
from useq._grid import (
    GridFromEdges,
    GridRowsColumns,
    GridWidthHeight,
    MultiPointPlan,
    RandomPoints,
    RelativeMultiPointPlan,
    Shape,
)
from useq._hardware_autofocus import AnyAutofocusPlan, AutoFocusPlan, AxesBasedAF
from useq._mda_event import MDAEvent, PropertyTuple
from useq._mda_sequence import MDASequence
from useq._plate import WellPlate, WellPlatePlan
from useq._plate_registry import register_well_plates, registered_well_plate_keys
from useq._point_visiting import OrderMode, TraversalOrder
from useq._position import AbsolutePosition, Position, RelativePosition
from useq._time import (
    AnyTimePlan,
    MultiPhaseTimePlan,
    TDurationLoops,
    TIntervalDuration,
    TIntervalLoops,
)
from useq._z import (
    AnyZPlan,
    ZAboveBelow,
    ZAbsolutePositions,
    ZRangeAround,
    ZRelativePositions,
    ZTopBottom,
)

__all__ = [
    "AbsolutePosition",
    "AcquireImage",
    "Action",
    "AnyAutofocusPlan",
    "AnyTimePlan",
    "AnyZPlan",
    "AutoFocusPlan",
    "AxesBasedAF",
    "Channel",
    "GridFromEdges",
    "GridRelative",
    "GridRowsColumns",
    "GridWidthHeight",
    "HardwareAutofocus",
    "MDAEvent",
    "MDASequence",
    "MultiPhaseTimePlan",
    "MultiPointPlan",
    "OrderMode",
    "Position",  # alias for AbsolutePosition
    "PropertyTuple",
    "RandomPoints",
    "register_well_plates",
    "registered_well_plate_keys",
    "RelativeMultiPointPlan",
    "RelativePosition",
    "Shape",
    "TDurationLoops",
    "TIntervalDuration",
    "TIntervalLoops",
    "TraversalOrder",
    "WellPlate",
    "WellPlatePlan",
    "ZAboveBelow",
    "ZAbsolutePositions",
    "ZRangeAround",
    "ZRelativePositions",
    "ZTopBottom",
]


MDAEvent.model_rebuild()
Position.model_rebuild()


def __getattr__(name: str) -> Any:
    if name == "GridRelative":
        from useq._grid import GridRowsColumns

        # warnings.warn(
        #     "useq.GridRelative has been renamed to useq.GridFromEdges",
        #     DeprecationWarning,
        #     stacklevel=2,
        # )

        return GridRowsColumns
    if name == "AnyGridPlan":  # pragma: no cover
        warnings.warn(
            "useq.AnyGridPlan has been renamed to useq.MultiPointPlan",
            DeprecationWarning,
            stacklevel=2,
        )
        return MultiPointPlan
    raise AttributeError(f"module {__name__} has no attribute {name}")



from __future__ import annotations

from typing import TYPE_CHECKING

from useq.pycromanager import to_pycromanager

if TYPE_CHECKING:
    from useq import MDASequence


def test_pycromanager_events(mda1: MDASequence) -> None:
    events = to_pycromanager(mda1)
    assert isinstance(events, list)
    assert isinstance(events[0], dict)



from typing import Any, Dict, Optional, Tuple

from pydantic import PrivateAttr

from useq._actions import HardwareAutofocus
from useq._base_model import FrozenModel
from useq._mda_event import MDAEvent


class AutoFocusPlan(FrozenModel):
    """Base class for hardware autofocus plans.

    Attributes
    ----------
    autofocus_device_name : str | None
        Optional name of the offset motor device.  If `None`, acquisition engines may
        attempt to set the offset however they see fit (such as using a current
        or default autofocus device.)
    autofocus_motor_offset : float | None
        Before autofocus is performed, the autofocus motor should be moved to this
        offset, if applicable. (Not all autofocus devices have an offset motor.)
        If None, the autofocus motor should not be moved.
    """

    autofocus_device_name: Optional[str] = None
    autofocus_motor_offset: Optional[float] = None

    def as_action(self) -> HardwareAutofocus:
        """Return a [`useq.HardwareAutofocus`][] for this autofocus plan."""
        return HardwareAutofocus(
            autofocus_device_name=self.autofocus_device_name,
            autofocus_motor_offset=self.autofocus_motor_offset,
        )

    def event(self, event: MDAEvent) -> Optional[MDAEvent]:
        """Return an autofocus [`useq.MDAEvent`][] if autofocus should be performed.

        The z position of the new [`useq.MDAEvent`][] is also updated if a relative
        zplan is provided since autofocus shuld be performed on the home z stack
        position.
        """
        if not self.should_autofocus(event):
            return None

        updates: Dict[str, Any] = {"action": self.as_action()}
        if event.z_pos is not None and event.sequence is not None:
            zplan = event.sequence.z_plan
            if zplan and zplan.is_relative and "z" in event.index:
                updates["z_pos"] = event.z_pos - list(zplan)[event.index["z"]]

        return event.model_copy(update=updates)

    def should_autofocus(self, event: MDAEvent) -> bool:
        """Method that must be implemented by a subclass.

        Should return True if autofocus should be performed (see
        [`useq.AxesBasedAF`][]).
        """
        raise NotImplementedError("should_autofocus() must be implemented by subclass.")


class AxesBasedAF(AutoFocusPlan):
    """Autofocus plan that performs autofocus when any of the specified axes change.

    Attributes
    ----------
    axes : Tuple[str, ...]
        Tuple of axis label to use for hardware autofocus.  At every event in which
        *any* axis in this tuple is change, autofocus will be performed.  For example,
        if `axes` is `('p',)` then autofocus will be performed every time the `p` axis
        is change, (in other words: every time the position is changed.).
    """

    axes: Tuple[str, ...]
    _previous: dict = PrivateAttr(default_factory=dict)

    def should_autofocus(self, event: MDAEvent) -> bool:
        """Return `True` if autofocus should be performed at this event.

        Will return `True` if any of the axes specified in `axes` have changed from the
        previous event.
        """
        self._previous, previous = dict(event.index), self._previous
        return any(
            axis in self.axes and previous.get(axis) != index
            for axis, index in event.index.items()
        )


AnyAutofocusPlan = AxesBasedAF



from typing import Any

import numpy as np
import pytest

import useq
from useq import _plate, _plate_registry


def test_plate_plan() -> None:
    pp = useq.WellPlatePlan(
        plate=96, a1_center_xy=(500, 200), rotation=5, selected_wells=np.s_[1:5:2, :6:3]
    )
    assert pp.plate.rows == 8
    assert pp.plate.columns == 12
    assert pp.plate.size == 96
    assert len(pp) == 4
    assert pp.selected_well_names == ["B1", "B4", "D1", "D4"]
    pos0 = next(iter(pp))
    assert isinstance(pos0, useq.Position)
    assert pos0 == pp[0]
    assert pos0.name == "B1"
    assert round(pos0.x, 2) == 1284.4  # first row, but rotataed slightly, µm
    assert round(pos0.y, 2) == -8765.75  # second column, µm

    js = pp.model_dump_json()

    pp2 = useq.WellPlatePlan.model_validate_json(js)
    assert pp2 == pp


def test_plate_plan_well_points() -> None:
    pp = useq.WellPlatePlan(
        plate=96,
        a1_center_xy=(500, 200),
        rotation="0.2rad",
        well_points_plan=useq.RandomPoints(num_points=10),
        selected_wells=slice(None),
    )
    assert len(pp) == 96 * 10

    pp2 = useq.WellPlatePlan(plate=96, a1_center_xy=(500, 200), selected_wells=None)
    assert len(pp2) == 0


def test_plate_plan_plot(monkeypatch: pytest.MonkeyPatch) -> None:
    mpl = pytest.importorskip("matplotlib.pyplot")
    monkeypatch.setattr(mpl, "show", lambda: None)
    pp = useq.WellPlatePlan(
        plate=96,
        a1_center_xy=(500, 200),
        rotation="4˚",
        selected_wells=np.s_[1:5:2, :6:3],
    )
    pp.plot()
    pp2 = useq.WellPlatePlan(plate=1536, a1_center_xy=(500, 200))
    pp2.plot()

    pp3 = useq.WellPlatePlan(
        plate=96,
        a1_center_xy=(500, 200),
        well_points_plan=useq.RandomPoints(num_points=10, fov_height=0.85, fov_width=1),
    )
    pp3.plot()


def test_plate_errors() -> None:
    useq.WellPlatePlan(plate=12, a1_center_xy=(0, 0), selected_wells=[(0, 0), (1, 2)])
    with pytest.raises(ValueError, match="Invalid well selection"):
        useq.WellPlatePlan(
            plate=12, a1_center_xy=(0, 0), selected_wells=[(0, 0), (100, 100)]
        )


def test_custom_plate(monkeypatch: pytest.MonkeyPatch) -> None:
    plates: dict = {}
    monkeypatch.setattr(_plate_registry, "_PLATE_REGISTRY", plates)
    monkeypatch.setattr(_plate, "_PLATE_REGISTRY", plates)

    useq.register_well_plates(
        {
            "silly": useq.WellPlate(
                rows=1, columns=1, well_spacing=1, circular_wells=False, well_size=1
            )
        },
        myplate={"rows": 8, "columns": 8, "well_spacing": 9, "well_size": 10},
    )
    assert "myplate" in plates
    assert "silly" in useq.registered_well_plate_keys()

    with pytest.raises(ValueError, match="Unknown plate name"):
        useq.WellPlatePlan(plate="bad", a1_center_xy=(0, 0))

    pp = useq.WellPlatePlan(plate="silly", a1_center_xy=(0, 0))
    assert pp.plate.rows == 1
    assert not pp.plate.circular_wells
    pp = useq.WellPlatePlan(plate="myplate", a1_center_xy=(0, 0))
    assert pp.plate.rows == 8


def test_plate_plan_serialization() -> None:
    pp = useq.WellPlatePlan(
        plate=96,
        a1_center_xy=(500, 200),
        rotation=5,
        selected_wells=np.s_[1:5:2, :6:3],
        well_points_plan=useq.RandomPoints(num_points=10),
    )
    js = pp.model_dump_json()
    pp2 = useq.WellPlatePlan.model_validate_json(js)
    assert pp2 == pp


def test_plate_plan_position_order() -> None:
    pp = useq.WellPlatePlan(
        plate=96,
        a1_center_xy=(0, 0),
        rotation=0,
        selected_wells=np.s_[1:5:2, :6:3],
        well_points_plan=useq.RandomPoints(num_points=3),
    )
    # check that the positions are ordered grouped by well name
    # e.g. A1_0000, A1_0001, B1_0000, B1_0001, ... and not A1_0000, B1_0000, A1_0001, B1_0001
    names = [p.name.split("_")[0] for p in pp]
    for i in range(0, len(names), 3):
        chunk = names[i : i + 3]
        assert len(set(chunk)) == 1, f"Chunk {chunk} does not have the same elements"


def test_plate_plan_equality() -> None:
    """Various ways of selecting wells should result in the same plan."""
    pp = useq.WellPlatePlan(
        plate=96, a1_center_xy=(0, 0), selected_wells=np.s_[1:5:2, :6:3]
    )
    pp2 = useq.WellPlatePlan(
        plate="96-well",
        a1_center_xy=(0, 0),
        selected_wells=[(1, 1, 3, 3), (0, 3, 0, 3)],
    )
    pp3 = useq.WellPlatePlan.model_validate_json(pp.model_dump_json())

    assert pp == pp2 == pp3


def test_plate_repr() -> None:
    # both can be reduced
    pp = useq.WellPlatePlan(
        plate=96, a1_center_xy=(0, 0), selected_wells=np.s_[1:5, 3:12:2]
    )
    rpp = repr(pp)
    assert "selected_wells=(slice(1, 5), slice(3, 12, 2))" in rpp
    assert eval(rpp, vars(useq)) == pp  # noqa: S307

    # can't be reduced
    pp = useq.WellPlatePlan(
        plate=96, a1_center_xy=(0, 0), selected_wells=[(1, 1, 1, 2), (7, 3, 4, 2)]
    )
    rpp = repr(pp)
    assert "selected_wells=((1, 1, 1, 2), (7, 3, 4, 2))" in rpp
    assert eval(rpp, vars(useq)) == pp  # noqa: S307

    # one can be reduced
    pp = useq.WellPlatePlan(
        plate=96, a1_center_xy=(0, 0), selected_wells=np.s_[(1, 2, 2, 3), 1:5]
    )
    rpp = repr(pp)
    assert "selected_wells=((1, 2, 2, 3), slice(1, 5))" in rpp
    assert eval(rpp, vars(useq)) == pp  # noqa: S307

    pp = useq.WellPlatePlan(plate=96, a1_center_xy=(0, 0), selected_wells=np.s_[:, 1:2])
    rpp = repr(pp)
    assert "selected_wells=(slice(8), slice(1, 2))" in rpp
    assert eval(rpp, vars(useq)) == pp  # noqa: S307


@pytest.mark.parametrize(
    "pp",
    [
        useq.WellPlatePlan(
            plate=96,
            a1_center_xy=(500, 200),
            rotation=5,
            selected_wells=np.s_[1:5:2, :6:3],
        ),
        {
            "plate": 96,
            "a1_center_xy": (500, 200),
            "rotation": 5,
            "selected_wells": np.s_[1:5:2, :6:3],
        },
    ],
)
def test_plate_plan_in_seq(pp: Any) -> None:
    seq = useq.MDASequence(stage_positions=pp)
    assert isinstance(seq.stage_positions, useq.WellPlatePlan)
    assert seq.stage_positions.plate.size == 96



from __future__ import annotations

from itertools import product
from typing import Any, Sequence

import useq
from useq import MDASequence

FITC = "FITC"
CY5 = "Cy5"
CY3 = "Cy3"
NAME = "name"
EMPTY: dict = {}
CH_FITC = useq.Channel(config=FITC, exposure=100)
CH_CY5 = useq.Channel(config=CY5, exposure=50)
Z_RANGE2 = useq.ZRangeAround(range=2, step=1)
Z_RANGE3 = useq.ZRangeAround(range=3, step=1)
Z_28_30 = useq.ZTopBottom(bottom=28, top=30, step=1)
Z_58_60 = useq.ZTopBottom(bottom=58, top=60, step=1)
GRID_2x2 = useq.GridRelative(rows=2, columns=2)
GRID_1100 = useq.GridFromEdges(top=1, bottom=-1, left=0, right=0)
GRID_2100 = useq.GridFromEdges(top=2, bottom=-1, left=0, right=0)
SEQ_1_CH = MDASequence(channels=[CH_FITC])
TLOOP2 = useq.TIntervalLoops(interval=1, loops=2)
TLOOP3 = useq.TIntervalLoops(interval=1, loops=3)
TLOOP5 = useq.TIntervalLoops(interval=1, loops=5)


def genindex(axes: dict[str, int]) -> list[dict[str, int]]:
    ranges = (range(x) for x in axes.values())
    return [dict(zip(axes, p)) for p in product(*ranges)]


def expect_mda(mda: MDASequence, **expectations: Sequence[Any]) -> None:
    results: dict[str, list[Any]] = {}
    for event in mda:
        for attr_name in expectations:
            results.setdefault(attr_name, []).append(getattr(event, attr_name))

    for attr_name, actual_value in results.items():
        assert actual_value == expectations[attr_name]


# test channels
def test_channel_only_in_position_sub_sequence() -> None:
    # test that a sub-position with a sequence has a channel, but not the main sequence
    expect_mda(
        MDASequence(stage_positions=[EMPTY, useq.Position(sequence=SEQ_1_CH)]),
        channel=[None, FITC],
        index=[{"p": 0}, {"p": 1, "c": 0}],
        exposure=[None, 100.0],
    )


def test_channel_in_main_and_position_sub_sequence() -> None:
    # test that a sub-position that specifies channel, overrides the global channel
    expect_mda(
        MDASequence(
            stage_positions=[EMPTY, useq.Position(sequence=SEQ_1_CH)],
            channels=[CH_CY5],
        ),
        channel=[CY5, FITC],
        index=[{"p": 0, "c": 0}, {"p": 1, "c": 0}],
        exposure=[50, 100.0],
    )


def test_subchannel_inherits_global_channel() -> None:
    # test that a sub-positions inherit the global channel
    mda = MDASequence(
        stage_positions=[EMPTY, {"sequence": {"z_plan": Z_28_30}}],
        channels=[CH_CY5],
    )
    assert all(e.channel.config == CY5 for e in mda)


# test grid_plan
def test_grid_relative_with_multi_stage_positions() -> None:
    # test that stage positions inherit the global relative grid plan

    expect_mda(
        MDASequence(
            stage_positions=[useq.Position(x=0, y=0), (10, 20)],
            grid_plan=GRID_2x2,
        ),
        index=genindex({"p": 2, "g": 4}),
        x_pos=[-0.5, 0.5, 0.5, -0.5, 9.5, 10.5, 10.5, 9.5],
        y_pos=[0.5, 0.5, -0.5, -0.5, 20.5, 20.5, 19.5, 19.5],
    )


def test_grid_relative_only_in_position_sub_sequence() -> None:
    # test a relative grid plan in a single stage position sub-sequence
    mda = MDASequence(
        stage_positions=[
            useq.Position(x=0, y=0),
            useq.Position(x=10, y=10, sequence={"grid_plan": GRID_2x2}),
        ],
    )

    expect_mda(
        mda,
        index=[
            {"p": 0},
            {"p": 1, "g": 0},
            {"p": 1, "g": 1},
            {"p": 1, "g": 2},
            {"p": 1, "g": 3},
        ],
        x_pos=[0.0, 9.5, 10.5, 10.5, 9.5],
        y_pos=[0.0, 10.5, 10.5, 9.5, 9.5],
    )


def test_grid_absolute_only_in_position_sub_sequence() -> None:
    # test a relative grid plan in a single stage position sub-sequence
    mda = MDASequence(
        stage_positions=[
            useq.Position(x=0, y=0),
            useq.Position(x=10, y=10, sequence={"grid_plan": GRID_1100}),
        ],
    )

    expect_mda(
        mda,
        index=[{"p": 0}, {"p": 1, "g": 0}, {"p": 1, "g": 1}, {"p": 1, "g": 2}],
        x_pos=[0.0, 0.0, 0.0, 0.0],
        y_pos=[0.0, 1.0, 0.0, -1.0],
    )


def test_grid_relative_in_main_and_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(x=0, y=0),
            useq.Position(name=NAME, x=10, y=10, sequence={"grid_plan": GRID_2x2}),
        ],
        grid_plan=GRID_2x2,
    )
    expect_mda(
        mda,
        index=genindex({"p": 2, "g": 4}),
        pos_name=[None] * 4 + [NAME] * 4,
        x_pos=[-0.5, 0.5, 0.5, -0.5, 9.5, 10.5, 10.5, 9.5],
        y_pos=[0.5, 0.5, -0.5, -0.5, 10.5, 10.5, 9.5, 9.5],
    )


def test_grid_absolute_in_main_and_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            EMPTY,
            useq.Position(name=NAME, sequence={"grid_plan": GRID_2100}),
        ],
        grid_plan=GRID_1100,
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "g": 0},
            {"p": 0, "g": 1},
            {"p": 0, "g": 2},
            {"p": 1, "g": 0},
            {"p": 1, "g": 1},
            {"p": 1, "g": 2},
            {"p": 1, "g": 3},
        ],
        pos_name=[None] * 3 + [NAME] * 4,
        x_pos=[0.0] * 7,
        y_pos=[1.0, 0.0, -1.0, 2.0, 1.0, 0.0, -1.0],
    )


def test_grid_absolute_in_main_and_grid_relative_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            EMPTY,
            useq.Position(name=NAME, x=10, y=10, sequence={"grid_plan": GRID_2x2}),
        ],
        grid_plan=GRID_1100,
    )

    expect_mda(
        mda,
        index=[
            {"p": 0, "g": 0},
            {"p": 0, "g": 1},
            {"p": 0, "g": 2},
            {"p": 1, "g": 0},
            {"p": 1, "g": 1},
            {"p": 1, "g": 2},
            {"p": 1, "g": 3},
        ],
        pos_name=[None] * 3 + [NAME] * 4,
        x_pos=[0.0, 0.0, 0.0, 9.5, 10.5, 10.5, 9.5],
        y_pos=[1.0, 0.0, -1.0, 10.5, 10.5, 9.5, 9.5],
    )


def test_grid_relative_in_main_and_grid_absolute_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(x=0, y=0),
            useq.Position(name=NAME, sequence={"grid_plan": GRID_1100}),
        ],
        grid_plan=GRID_2x2,
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "g": 0},
            {"p": 0, "g": 1},
            {"p": 0, "g": 2},
            {"p": 0, "g": 3},
            {"p": 1, "g": 0},
            {"p": 1, "g": 1},
            {"p": 1, "g": 2},
        ],
        pos_name=[None] * 4 + [NAME] * 3,
        x_pos=[-0.5, 0.5, 0.5, -0.5, 0.0, 0.0, 0.0],
        y_pos=[0.5, 0.5, -0.5, -0.5, 1.0, 0.0, -1.0],
    )


def test_multi_g_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            {"sequence": {"grid_plan": {"rows": 1, "columns": 2}}},
            {"sequence": {"grid_plan": GRID_2x2}},
            {"sequence": {"grid_plan": GRID_1100}},
        ]
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "g": 0},
            {"p": 0, "g": 1},
            {"p": 1, "g": 0},
            {"p": 1, "g": 1},
            {"p": 1, "g": 2},
            {"p": 1, "g": 3},
            {"p": 2, "g": 0},
            {"p": 2, "g": 1},
            {"p": 2, "g": 2},
        ],
        x_pos=[-0.5, 0.5, -0.5, 0.5, 0.5, -0.5, 0.0, 0.0, 0.0],
        y_pos=[0.0, 0.0, 0.5, 0.5, -0.5, -0.5, 1.0, 0.0, -1.0],
    )


# test_z_plan
def test_z_relative_with_multi_stage_positions() -> None:
    expect_mda(
        mda=MDASequence(stage_positions=[(0, 0, 0), (10, 20, 10)], z_plan=Z_RANGE2),
        index=genindex({"p": 2, "z": 3}),
        x_pos=[0.0, 0.0, 0.0, 10.0, 10.0, 10.0],
        y_pos=[0.0, 0.0, 0.0, 20.0, 20.0, 20.0],
        z_pos=[-1.0, 0.0, 1.0, 9.0, 10.0, 11.0],
    )


def test_z_absolute_with_multi_stage_positions() -> None:
    expect_mda(
        MDASequence(
            stage_positions=[useq.Position(x=0, y=0), (10, 20)], z_plan=Z_58_60
        ),
        index=genindex({"p": 2, "z": 3}),
        x_pos=[0.0, 0.0, 0.0, 10.0, 10.0, 10.0],
        y_pos=[0.0, 0.0, 0.0, 20.0, 20.0, 20.0],
        z_pos=[58.0, 59.0, 60.0, 58.0, 59.0, 60.0],
    )


def test_z_relative_only_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(z=0),
            useq.Position(name=NAME, z=10, sequence={"z_plan": Z_RANGE2}),
        ],
    )

    expect_mda(
        mda,
        index=[{"p": 0}, {"p": 1, "z": 0}, {"p": 1, "z": 1}, {"p": 1, "z": 2}],
        pos_name=[None, NAME, NAME, NAME],
        z_pos=[0.0, 9.0, 10.0, 11.0],
    )


def test_z_absolute_only_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(z=0),
            useq.Position(name=NAME, sequence={"z_plan": Z_58_60}),
        ],
    )

    expect_mda(
        mda,
        index=[{"p": 0}, {"p": 1, "z": 0}, {"p": 1, "z": 1}, {"p": 1, "z": 2}],
        pos_name=[None, NAME, NAME, NAME],
        z_pos=[0.0, 58, 59, 60],
    )


def test_z_relative_in_main_and_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(z=0),
            useq.Position(name=NAME, z=10, sequence={"z_plan": Z_RANGE3}),
        ],
        z_plan=Z_RANGE2,
    )

    indices = genindex({"p": 2, "z": 4})
    indices.pop(3)
    expect_mda(
        mda,
        index=indices,
        pos_name=[None] * 3 + [NAME] * 4,
        z_pos=[-1.0, 0.0, 1.0, 8.5, 9.5, 10.5, 11.5],
    )


def test_z_absolute_in_main_and_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            EMPTY,
            useq.Position(name=NAME, sequence={"z_plan": Z_28_30}),
        ],
        z_plan=Z_58_60,
    )
    expect_mda(
        mda,
        index=genindex({"p": 2, "z": 3}),
        pos_name=[None] * 3 + [NAME] * 3,
        z_pos=[58.0, 59.0, 60.0, 28.0, 29.0, 30.0],
    )


def test_z_absolute_in_main_and_z_relative_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            EMPTY,
            useq.Position(name=NAME, z=10, sequence={"z_plan": Z_RANGE3}),
        ],
        z_plan=Z_58_60,
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "z": 0},
            {"p": 0, "z": 1},
            {"p": 0, "z": 2},
            {"p": 1, "z": 0},
            {"p": 1, "z": 1},
            {"p": 1, "z": 2},
            {"p": 1, "z": 3},
        ],
        pos_name=[None] * 3 + [NAME] * 4,
        z_pos=[58.0, 59.0, 60.0, 8.5, 9.5, 10.5, 11.5],
    )


def test_z_relative_in_main_and_z_absolute_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            useq.Position(z=0),
            useq.Position(name=NAME, sequence={"z_plan": Z_58_60}),
        ],
        z_plan=Z_RANGE3,
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "z": 0},
            {"p": 0, "z": 1},
            {"p": 0, "z": 2},
            {"p": 0, "z": 3},
            {"p": 1, "z": 0},
            {"p": 1, "z": 1},
            {"p": 1, "z": 2},
        ],
        pos_name=[None] * 4 + [NAME] * 3,
        z_pos=[-1.5, -0.5, 0.5, 1.5, 58.0, 59.0, 60.0],
    )


def test_multi_z_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[
            {"sequence": {"z_plan": Z_58_60}},
            {"sequence": {"z_plan": Z_RANGE3}},
            {"sequence": {"z_plan": Z_28_30}},
        ],
    )
    expect_mda(
        mda,
        index=[
            {"p": 0, "z": 0},
            {"p": 0, "z": 1},
            {"p": 0, "z": 2},
            {"p": 1, "z": 0},
            {"p": 1, "z": 1},
            {"p": 1, "z": 2},
            {"p": 1, "z": 3},
            {"p": 2, "z": 0},
            {"p": 2, "z": 1},
            {"p": 2, "z": 2},
        ],
        z_pos=[58.0, 59.0, 60.0, -1.5, -0.5, 0.5, 1.5, 28.0, 29.0, 30.0],
    )


# test time_plan
def test_t_with_multi_stage_positions() -> None:
    expect_mda(
        MDASequence(stage_positions=[EMPTY, EMPTY], time_plan=[TLOOP2]),
        index=genindex({"t": 2, "p": 2}),
        min_start_time=[0.0, 0.0, 1.0, 1.0],
    )


def test_t_only_in_position_sub_sequence() -> None:
    expect_mda(
        MDASequence(stage_positions=[EMPTY, {"sequence": {"time_plan": [TLOOP5]}}]),
        index=[
            {"p": 0},
            {"p": 1, "t": 0},
            {"p": 1, "t": 1},
            {"p": 1, "t": 2},
            {"p": 1, "t": 3},
            {"p": 1, "t": 4},
        ],
        min_start_time=[None, 0.0, 1.0, 2.0, 3.0, 4.0],
    )


def test_t_in_main_and_in_position_sub_sequence() -> None:
    mda = MDASequence(
        stage_positions=[EMPTY, {"sequence": {"time_plan": [TLOOP5]}}],
        time_plan=[TLOOP2],
    )
    expect_mda(
        mda,
        index=[
            {"t": 0, "p": 0},
            {"t": 0, "p": 1},
            {"t": 1, "p": 1},
            {"t": 2, "p": 1},
            {"t": 3, "p": 1},
            {"t": 4, "p": 1},
            {"t": 1, "p": 0},
            {"t": 0, "p": 1},
            {"t": 1, "p": 1},
            {"t": 2, "p": 1},
            {"t": 3, "p": 1},
            {"t": 4, "p": 1},
        ],
        min_start_time=[0.0, 0.0, 1.0, 2.0, 3.0, 4.0, 1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
    )


def test_mix_cgz_axes() -> None:
    mda = MDASequence(
        axis_order="tpgcz",
        stage_positions=[
            useq.Position(x=0, y=0),
            useq.Position(
                name=NAME,
                x=10,
                y=10,
                z=30,
                sequence=MDASequence(
                    channels=[
                        {"config": FITC, "exposure": 200},
                        {"config": CY3, "exposure": 100},
                    ],
                    grid_plan=useq.GridRelative(rows=2, columns=1),
                    z_plan=Z_RANGE2,
                ),
            ),
        ],
        channels=[CH_CY5],
        z_plan={"top": 100, "bottom": 98, "step": 1},
        grid_plan=GRID_1100,
    )
    expect_mda(
        mda,
        index=[
            *genindex({"p": 1, "g": 3, "c": 1, "z": 3}),
            {"p": 1, "g": 0, "c": 0, "z": 0},
            {"p": 1, "g": 0, "c": 0, "z": 1},
            {"p": 1, "g": 0, "c": 0, "z": 2},
            {"p": 1, "g": 0, "c": 1, "z": 0},
            {"p": 1, "g": 0, "c": 1, "z": 1},
            {"p": 1, "g": 0, "c": 1, "z": 2},
            {"p": 1, "g": 1, "c": 0, "z": 0},
            {"p": 1, "g": 1, "c": 0, "z": 1},
            {"p": 1, "g": 1, "c": 0, "z": 2},
            {"p": 1, "g": 1, "c": 1, "z": 0},
            {"p": 1, "g": 1, "c": 1, "z": 1},
            {"p": 1, "g": 1, "c": 1, "z": 2},
        ],
        pos_name=[None] * 9 + [NAME] * 12,
        x_pos=[0.0] * 9 + [10.0] * 12,
        y_pos=[1, 1, 1, 0, 0, 0, -1, -1, -1] + [10.5] * 6 + [9.5] * 6,
        z_pos=[98.0, 99.0, 100.0] * 3 + [29.0, 30.0, 31.0] * 4,
        channel=[CY5] * 9 + ([FITC] * 3 + [CY3] * 3) * 2,
        exposure=[50.0] * 9 + [200.0] * 3 + [100.0] * 3 + [200.0] * 3 + [100.0] * 3,
    )


# axes order????
def test_order() -> None:
    sub_pos = useq.Position(
        z=50,
        sequence=MDASequence(
            channels=[CH_FITC, useq.Channel(config=CY3, exposure=200)]
        ),
    )
    mda = MDASequence(
        stage_positions=[useq.Position(z=0), sub_pos],
        channels=[CH_FITC, CH_CY5],
        z_plan=useq.ZRangeAround(range=2, step=1),
    )

    # might appear confusing at first, but the sub-position had no z plan to iterate
    # so, specifying a different axis_order for the subplan does not change the
    # order of the z positions (specified globally)
    expected_indices = [
        {"p": 0, "c": 0, "z": 0},
        {"p": 0, "c": 0, "z": 1},
        {"p": 0, "c": 0, "z": 2},
        {"p": 0, "c": 1, "z": 0},
        {"p": 0, "c": 1, "z": 1},
        {"p": 0, "c": 1, "z": 2},
        {"p": 1, "c": 0, "z": 0},
        {"p": 1, "c": 1, "z": 0},
        {"p": 1, "c": 0, "z": 1},
        {"p": 1, "c": 1, "z": 1},
        {"p": 1, "c": 0, "z": 2},
        {"p": 1, "c": 1, "z": 2},
    ]
    expect_pos = [-1.0, 0.0, 1.0, -1.0, 0.0, 1.0, 49.0, 49.0, 50.0, 50.0, 51.0, 51.0]
    expect_ch = [FITC] * 3 + [CY5] * 3 + [FITC, CY3] * 3
    expect_mda(mda, index=expected_indices, z_pos=expect_pos, channel=expect_ch)


def test_channels_and_pos_grid_plan() -> None:
    # test that all channels are acquired for each grid position
    sub_seq = MDASequence(grid_plan=useq.GridRelative(rows=2, columns=1))
    mda = MDASequence(
        channels=[CH_CY5, CH_FITC],
        stage_positions=[useq.Position(x=0, y=0, sequence=sub_seq)],
    )

    expect_mda(
        mda,
        index=genindex({"p": 1, "c": 2, "g": 2}),
        x_pos=[0.0, 0.0, 0.0, 0.0],
        y_pos=[0.5, -0.5, 0.5, -0.5],
        channel=[CY5, CY5, FITC, FITC],
    )


def test_channels_and_pos_z_plan() -> None:
    # test that all channels are acquired for each z position
    mda = MDASequence(
        channels=[CH_CY5, CH_FITC],
        stage_positions=[useq.Position(x=0, y=0, z=0, sequence={"z_plan": Z_RANGE2})],
    )
    expect_mda(
        mda,
        index=genindex({"p": 1, "c": 2, "z": 3}),
        z_pos=[-1.0, 0.0, 1.0, -1.0, 0.0, 1.0],
        channel=[CY5, CY5, CY5, FITC, FITC, FITC],
    )


def test_channels_and_pos_time_plan() -> None:
    # test that all channels are acquired for each timepoint
    mda = MDASequence(
        axis_order="tpgcz",
        channels=[CH_CY5, CH_FITC],
        stage_positions=[useq.Position(x=0, y=0, sequence={"time_plan": [TLOOP3]})],
    )
    expect_mda(
        mda,
        index=genindex({"p": 1, "c": 2, "t": 3}),
        min_start_time=[0.0, 1.0, 2.0, 0.0, 1.0, 2.0],
        channel=[CY5, CY5, CY5, FITC, FITC, FITC],
    )


def test_channels_and_pos_z_grid_and_time_plan() -> None:
    # test that all channels are acquired for each z and grid positions
    sub_seq = useq.MDASequence(grid_plan=GRID_2x2, z_plan=Z_RANGE2, time_plan=[TLOOP2])
    mda = MDASequence(
        channels=[CH_CY5, CH_FITC],
        stage_positions=[useq.Position(x=0, y=0, sequence=sub_seq)],
    )

    expect_mda(mda, channel=[CY5] * 24 + [FITC] * 24)


def test_sub_channels_and_any_plan() -> None:
    # test that only specified sub-channels are acquired for each z plan
    mda = MDASequence(
        channels=[CY5, FITC],
        stage_positions=[{"sequence": {"channels": [FITC], "z_plan": Z_RANGE2}}],
    )

    expect_mda(mda, channel=[FITC, FITC, FITC])



# don't add __future__.annotations here
# pydantic2 isn't rebuilding the model correctly

from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)

from pydantic import Field, field_validator

from useq._actions import AcquireImage, AnyAction
from useq._base_model import UseqModel

try:
    from pydantic import field_serializer
except ImportError:
    field_serializer = None  # type: ignore

if TYPE_CHECKING:
    from useq._mda_sequence import MDASequence

    ReprArgs = Sequence[Tuple[Optional[str], Any]]


class Channel(UseqModel):
    """Channel in a MDA event.

    Attributes
    ----------
    config : str
        Name of the configuration to use for this channel, (e.g. `"488nm"`, `"DAPI"`,
        `"FITC"`).
    group : str
        Optional name of the group to which this channel belongs. By default,
        `"Channel"`.
    """

    config: str
    group: str = "Channel"

    def __eq__(self, _value: object) -> bool:
        if isinstance(_value, str):
            return self.config == _value
        return super().__eq__(_value)


class PropertyTuple(NamedTuple):
    """Three-tuple capturing a device, property, and value.

    Attributes
    ----------
    device_name : str
        Name of a device.
    property_name : str
        Name of a property recognized by the device.
    value : Any
        Value for the property.
    """

    device_name: str
    property_name: str
    property_value: Any


def _float_or_none(v: Any) -> Optional[float]:
    return float(v) if v is not None else v


class MDAEvent(UseqModel):
    """Define a single event in a [`MDASequence`][useq.MDASequence].

    Usually, this object will be generator by iterating over a
    [`MDASequence`][useq.MDASequence] (see [`useq.MDASequence.iter_events`][]).

    Attributes
    ----------
    index : dict[str, int]
        Index of this event in the sequence. This is a mapping of axis name
        to index.  For example: `{'t': 4, 'c': 0, 'z': 5},`
    channel : Channel | None
        Channel to use for this event. If `None`, implies use current channel.
        By default, `None`.  `Channel` is a simple pydantic object with two attributes:
        `config` and `group`.  `config` is the name of the configuration to use for this
        channel, (e.g. `"488nm"`, `"DAPI"`, `"FITC"`).  `group` is the name of the group
        to which this channel belongs. By default, `"Channel"`.
    exposure : float | None
        Exposure time in milliseconds. If not provided, implies use current exposure
        time. By default, `None`.
    min_start_time : float | None
        Minimum start time of this event, in seconds.  If provided, the engine will
        pause until this time has elapsed before starting this event. Times are
        relative to the start of the sequence, or the last event with
        `reset_event_timer` set to `True`.
    pos_name : str | None
        The name assigned to the position. By default, `None`.
    x_pos : float | None
        X position in microns. If not provided, implies use current position. By
        default, `None`.
    y_pos : float | None
        Y position in microns. If not provided, implies use current position. By
        default, `None`.
    z_pos : float | None
        Z position in microns. If not provided, implies use current position. By
        default, `None`.
    sequence : MDASequence | None
        A reference to the [`useq.MDASequence`][] this event belongs to. This is a
        read-only attribute. By default, `None`.
    properties : Sequence[PropertyTuple] | None
        List of [`useq.PropertyTuple`][] to set before starting this event. Where each
        item in the list is a 3-member named tuple of `(device_name, property_name,
        property_value)`.  This is inspired by micro-manager's Device Adapter API, but
        could be used to set arbitrary properties in any backend that supports the
        concept of devices that have properties with values. By default, `None`.
    metadata : dict
        Optional metadata to be associated with this event.
    action : Action
        The action to perform for this event.  By default, [`useq.AcquireImage`][].
        Example of another action is [`useq.HardwareAutofocus`][] which could be used
        to perform a hardware autofocus.
    keep_shutter_open : bool
        If `True`, the illumination shutter should be left open after the event has
        been executed, otherwise it should be closed. By default, `False`."
        This is useful when the sequence of events being executed use the same
        illumination scheme (such as a z-stack in a single channel), and closing and
        opening the shutter between events would be slow.
    reset_event_timer : bool
        If `True`, the engine should reset the event timer to the time of this event,
        and future `min_start_time` values will be relative to this event. By default,
        `False`.
    """

    # MappingProxyType is not subscriptable on Python 3.8
    index: Mapping[str, int] = Field(default_factory=lambda: MappingProxyType({}))
    channel: Optional[Channel] = None
    exposure: Optional[float] = Field(default=None, gt=0.0)
    min_start_time: Optional[float] = None  # time in sec
    pos_name: Optional[str] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None
    z_pos: Optional[float] = None
    sequence: Optional["MDASequence"] = Field(default=None, repr=False)
    properties: Optional[List[PropertyTuple]] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)
    action: AnyAction = Field(default_factory=AcquireImage)
    keep_shutter_open: bool = False
    reset_event_timer: bool = False

    @field_validator("channel", mode="before")
    def _validate_channel(cls, val: Any) -> Any:
        return Channel(config=val) if isinstance(val, str) else val

    if field_serializer is not None:
        _si = field_serializer("index", mode="plain")(lambda v: dict(v))
        _sx = field_serializer("x_pos", mode="plain")(_float_or_none)
        _sy = field_serializer("y_pos", mode="plain")(_float_or_none)
        _sz = field_serializer("z_pos", mode="plain")(_float_or_none)



from __future__ import annotations

from functools import lru_cache
from itertools import product
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Iterator, cast

from typing_extensions import TypedDict

from useq._channel import Channel  # noqa: TCH001  # noqa: TCH001
from useq._mda_event import Channel as EventChannel
from useq._mda_event import MDAEvent
from useq._utils import AXES, Axis, _has_axes
from useq._z import AnyZPlan  # noqa: TCH001  # noqa: TCH001

if TYPE_CHECKING:
    from useq._mda_sequence import MDASequence
    from useq._position import Position, PositionBase, RelativePosition


class MDAEventDict(TypedDict, total=False):
    index: MappingProxyType[str, int]
    channel: EventChannel | None
    exposure: float | None
    min_start_time: float | None
    pos_name: str | None
    x_pos: float | None
    y_pos: float | None
    z_pos: float | None
    sequence: MDASequence | None
    # properties: list[tuple] | None
    metadata: dict
    reset_event_timer: bool


class PositionDict(TypedDict, total=False):
    x_pos: float
    y_pos: float
    z_pos: float


@lru_cache(maxsize=None)
def _iter_axis(seq: MDASequence, ax: str) -> tuple[Channel | float | PositionBase, ...]:
    return tuple(seq.iter_axis(ax))


@lru_cache(maxsize=None)
def _sizes(seq: MDASequence) -> dict[str, int]:
    return {k: len(list(_iter_axis(seq, k))) for k in seq.axis_order}


@lru_cache(maxsize=None)
def _used_axes(seq: MDASequence) -> str:
    return "".join(k for k in seq.axis_order if _sizes(seq)[k])


def iter_sequence(sequence: MDASequence) -> Iterator[MDAEvent]:
    """Iterate over all events in the MDA sequence.'.

    !!! note
        This method will usually be used via [`useq.MDASequence.iter_events`][], or by
        simply iterating over the sequence.

    This does the job of iterating over all the frames in the MDA sequence,
    handling the logic of merging all z plans in channels and stage positions
    defined in the plans for each axis.

    The is the most "logic heavy" part of `useq-schema` (the rest of which is
    almost entirely declarative).  This iterator is useful for consuming `MDASequence`
    objects in a python runtime, but it isn't considered a "core" part of the schema.

    Parameters
    ----------
    sequence : MDASequence
        The sequence to iterate over.

    Yields
    ------
    MDAEvent
        Each event in the MDA sequence.
    """
    if not (keep_shutter_open_axes := sequence.keep_shutter_open_across):
        yield from _iter_sequence(sequence)
        return

    it = _iter_sequence(sequence)
    if (this_e := next(it, None)) is None:  # pragma: no cover
        return

    for next_e in it:
        # set `keep_shutter_open` to `True` if and only if ALL axes whose index
        # changes betwee this_event and next_event are in `keep_shutter_open_axes`
        if all(
            axis in keep_shutter_open_axes
            for axis, idx in this_e.index.items()
            if idx != next_e.index[axis]
        ):
            this_e = this_e.model_copy(update={"keep_shutter_open": True})
        yield this_e
        this_e = next_e
    yield this_e


def _iter_sequence(
    sequence: MDASequence,
    *,
    base_event_kwargs: MDAEventDict | None = None,
    event_kwarg_overrides: MDAEventDict | None = None,
    position_offsets: PositionDict | None = None,
    _last_t_idx: int = -1,
) -> Iterator[MDAEvent]:
    """Helper function for `iter_sequence`.

    We put most of the logic into this sub-function so that `iter_sequence` can
    easily modify the resulting sequence of events (e.g. to peek at the next event
    before yielding the current one).

    It also keeps the sub-sequence iteration kwargs out of the public API.

    Parameters
    ----------
    sequence : MDASequence
        The sequence to iterate over.
    base_event_kwargs : MDAEventDict | None
        A dictionary of "global" kwargs to begin with when building the kwargs passed
        to each MDAEvent.  These will be overriden by event-specific kwargs (e.g. if
        the event specifies a channel, it will be used instead of the
        `base_event_kwargs`.)
    event_kwarg_overrides : MDAEventDict | None
        A dictionary of kwargs that will be applied to all events. Unlike
        `base_event_kwargs`, these kwargs take precedence over any event-specific
        kwargs.
    position_offsets : PositionDict | None
        A dictionary of offsets to apply to each position. This can be used to shift
        all positions in a sub-sequence.  Keys must be one of `x_pos`, `y_pos`, or
        `z_pos` and values should be floats.s
    _last_t_idx : int
        The index of the last timepoint.  This is used to determine if the event
        should reset the event timer.

    Yields
    ------
    MDAEvent
        Each event in the MDA sequence.
    """
    order = _used_axes(sequence)
    # this needs to be tuple(...) to work for mypyc
    axis_iterators = tuple(enumerate(_iter_axis(sequence, ax)) for ax in order)
    for item in product(*axis_iterators):
        if not item:  # the case with no events
            continue  # pragma: no cover
        # get axes objects for this event
        index, time, position, grid, channel, z_pos = _parse_axes(zip(order, item))

        # skip if necessary
        if _should_skip(position, channel, index, sequence.z_plan):
            continue

        # build kwargs that will be passed to this MDAEvent
        event_kwargs = base_event_kwargs or MDAEventDict(sequence=sequence)
        # the .update() here lets us build on top of the base_event.index if present

        event_kwargs["index"] = MappingProxyType(
            {**event_kwargs.get("index", {}), **index}  # type: ignore
        )
        # determine x, y, z positions
        event_kwargs.update(_xyzpos(position, channel, sequence.z_plan, grid, z_pos))
        if position and position.name:
            event_kwargs["pos_name"] = position.name
        if channel:
            event_kwargs["channel"] = EventChannel.model_construct(
                config=channel.config, group=channel.group
            )
            if channel.exposure is not None:
                event_kwargs["exposure"] = channel.exposure
        if time is not None:
            event_kwargs["min_start_time"] = time

        # apply any overrides
        if event_kwarg_overrides:
            event_kwargs.update(event_kwarg_overrides)

        # shift positions if position_offsets have been provided
        # (usually from sub-sequences)
        if position_offsets:
            for k, v in position_offsets.items():
                if event_kwargs[k] is not None:  # type: ignore[literal-required]
                    event_kwargs[k] += v  # type: ignore[literal-required]

        # grab global autofocus plan (may be overridden by position-specific plan below)
        autofocus_plan = sequence.autofocus_plan

        # if a position has been declared with a sub-sequence, we recurse into it
        if position:
            if _has_axes(position.sequence):
                # determine any relative position shifts or global overrides
                _pos, _offsets = _position_offsets(position, event_kwargs)
                # build overrides for this position
                pos_overrides = MDAEventDict(sequence=sequence, **_pos)
                if position.name:
                    pos_overrides["pos_name"] = position.name

                sub_seq = position.sequence
                # if the sub-sequence doe not have an autofocus plan, we override it
                # with the parent sequence's autofocus plan
                if not sub_seq.autofocus_plan:
                    sub_seq = sub_seq.model_copy(
                        update={"autofocus_plan": autofocus_plan}
                    )

                # recurse into the sub-sequence
                yield from _iter_sequence(
                    sub_seq,
                    base_event_kwargs=event_kwargs.copy(),
                    event_kwarg_overrides=pos_overrides,
                    position_offsets=_offsets,
                    _last_t_idx=_last_t_idx,
                )
                continue
            # note that position.sequence may be Falsey even if not None, for example
            # if all it has is an autofocus plan.  In that case, we don't recurse.
            # and we don't hit the continue statement, but we can use the autofocus plan
            elif position.sequence is not None and position.sequence.autofocus_plan:
                autofocus_plan = position.sequence.autofocus_plan

        if event_kwargs["index"].get(Axis.TIME) == 0 and _last_t_idx != 0:
            event_kwargs["reset_event_timer"] = True
        event = MDAEvent.model_construct(**event_kwargs)
        if autofocus_plan:
            af_event = autofocus_plan.event(event)
            if af_event:
                yield af_event
        yield event
        _last_t_idx = event.index.get(Axis.TIME, _last_t_idx)


# ###################### Helper functions ######################


def _position_offsets(
    position: Position, event_kwargs: MDAEventDict
) -> tuple[MDAEventDict, PositionDict]:
    """Determine shifts and position overrides for position subsequences."""
    pos_seq = cast("MDASequence", position.sequence)
    overrides = MDAEventDict()
    offsets = PositionDict()
    if not pos_seq.z_plan:
        # if this position has no z_plan, we use the z_pos from the parent
        overrides["z_pos"] = event_kwargs.get("z_pos")
    elif pos_seq.z_plan.is_relative:
        # otherwise apply z-shifts if this position has a relative z_plan
        offsets["z_pos"] = position.z or 0.0

    if not pos_seq.grid_plan:
        # if this position has no grid plan, we use the x_pos and y_pos from the parent
        overrides["x_pos"] = event_kwargs.get("x_pos")
        overrides["y_pos"] = event_kwargs.get("y_pos")
    elif pos_seq.grid_plan.is_relative:
        # otherwise apply x/y shifts if this position has a relative grid plan
        offsets["x_pos"] = position.x or 0.0
        offsets["y_pos"] = position.y or 0.0
    return overrides, offsets


def _parse_axes(
    event: zip[tuple[str, Any]],
) -> tuple[
    dict[str, int],
    float | None,  # time
    Position | None,
    RelativePosition | None,
    Channel | None,
    float | None,  # z
]:
    """Parse an individual event from the product of axis iterators.

    Returns typed objects for each axis, and the index of the event.
    """
    # NOTE: this is currently the biggest time sink in iter_sequence.
    # It is called for every event and takes ~40% of the cumulative time.
    _ev = dict(event)
    index = {ax: _ev[ax][0] for ax in AXES if ax in _ev}
    # this needs to be tuple(...) to work for mypyc
    axes = tuple(_ev[ax][1] if ax in _ev else None for ax in AXES)
    return (index, *axes)


def _should_skip(
    position: Position | None,
    channel: Channel | None,
    index: dict[str, int],
    z_plan: AnyZPlan | None,
) -> bool:
    """Return True if this event should be skipped."""
    if channel:
        # skip channels
        if Axis.TIME in index and index[Axis.TIME] % channel.acquire_every:
            return True

        # only acquire on the middle plane:
        if (
            not channel.do_stack
            and z_plan is not None
            and index[Axis.Z] != z_plan.num_positions() // 2
        ):
            return True

    if (
        not position
        or position.sequence is None
        or position.sequence.autofocus_plan is not None
    ):
        return False

    # NOTE: if we ever add more plans, they will need to be explicitly added
    # https://github.com/pymmcore-plus/useq-schema/pull/85

    # get if sub-sequence has any plan
    plans = any(
        (
            position.sequence.grid_plan,
            position.sequence.z_plan,
            position.sequence.time_plan,
        )
    )
    # overwriting the *global* channel index since it is no longer relevant.
    # if channel IS SPECIFIED in the position.sequence WITH any plan,
    # we skip otherwise the channel will be acquired twice. Same happens if
    # the channel IS NOT SPECIFIED but ANY plan is.
    if index.get(Axis.CHANNEL, 0) != 0:
        if (position.sequence.channels and plans) or not plans:
            return True
    if Axis.Z in index and index[Axis.Z] != 0 and position.sequence.z_plan:
        return True
    if Axis.GRID in index and index[Axis.GRID] != 0 and position.sequence.grid_plan:
        return True
    return False


def _xyzpos(
    position: Position | None,
    channel: Channel | None,
    z_plan: AnyZPlan | None,
    grid: RelativePosition | None = None,
    z_pos: float | None = None,
) -> MDAEventDict:
    if z_pos is not None:
        # combine z_pos with z_offset
        if channel and channel.z_offset is not None:
            z_pos += channel.z_offset
        if z_plan and z_plan.is_relative:
            # TODO: either disallow without position z, or add concept of "current"
            z_pos += getattr(position, Axis.Z, None) or 0
    elif position:
        z_pos = position.z

    if grid:
        x_pos: float | None = grid.x
        y_pos: float | None = grid.y
        if grid.is_relative:
            px = getattr(position, "x", 0) or 0
            py = getattr(position, "y", 0) or 0
            x_pos = x_pos + px if x_pos is not None else None
            y_pos = y_pos + py if y_pos is not None else None
    else:
        x_pos = getattr(position, "x", None)
        y_pos = getattr(position, "y", None)

    return {"x_pos": x_pos, "y_pos": y_pos, "z_pos": z_pos}



from typing import Optional, Union

from typing_extensions import Literal

from useq._base_model import FrozenModel


class Action(FrozenModel):
    """Base class for a [`useq.MDAEvent`][] action.

    An `Action` specifies what task should be performed during a
    [`useq.MDAEvent`][]. An `Action` can be for example used to acquire an
    image ([`useq.AcquireImage`][]) or to perform a hardware autofocus
    ([`useq.HardwareAutofocus`][]).

    Attributes
    ----------
    type : str
        Type of the action that should be performed at the [`useq.MDAEvent`][].
    """

    type: str


class AcquireImage(Action):
    """[`useq.Action`][] to acquire an image.

    Attributes
    ----------
    type : Literal["acquire_image"]
        This action can be used to acquire an image.
    """

    type: Literal["acquire_image"] = "acquire_image"


class HardwareAutofocus(Action):
    """[`useq.Action`][] to perform a hardware autofocus.

    See also [`useq.AutoFocusPlan`][].

    Attributes
    ----------
    type : Literal["hardware_autofocus"]
        This action can be used to trigger hardware autofocus.
    autofocus_device_name : str, optional
        The name of the autofocus offset motor device (if applicable).  If `None`,
        acquisition engines may attempt to set the offset however they see fit (such as
        using a current or default autofocus device.)
    autofocus_motor_offset: float, optional
        Before autofocus is performed, the autofocus motor should be moved to this
        offset, if applicable. (Not all autofocus devices have an offset motor.)
        If None, the autofocus motor should not be moved.
    max_retries : int
        The number of retries if autofocus fails. By default, 3.
    """

    type: Literal["hardware_autofocus"] = "hardware_autofocus"
    autofocus_device_name: Optional[str] = None
    autofocus_motor_offset: Optional[float] = None
    max_retries: int = 3


AnyAction = Union[HardwareAutofocus, AcquireImage]
